# Copyright (C) 2019-2023 Columbia Experimental Gravity Group (GECo)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# pylint: disable=line-too-long,invalid-name,bad-continuation
# flake8: noqa

"""
A partial HEALPix skymap class supporting multi-resolution HEALPix skymaps in
NUNIQ ordering.
"""

import sys
import functools
import operator
import base64
from io import BytesIO
from textwrap import wrap, dedent
from collections import OrderedDict
from typing import (
    List,
    Iterator,
    MutableMapping,
    Callable,
    Literal,
    Sequence,
    Optional,
    Union,
    IO,
    Tuple,
    Any,
    Type,
    TypeVar,
    TYPE_CHECKING,
    cast,
    overload
)
from .utils import (
    max_uint_type,
    uniq2nest,
    uniq2dangle,
    uniq_diadic,
    uniq_intersection,
    fill,
    render,
    reraster,
    check_valid_nuniq,
    uniq2nest_and_nside,
    uniq2nside,
    uniq2order,
    nest2ang,
    nside2pixarea,
    nside_quantile_indices,
    nside_slices,
    nest2uniq,
    uniq_minimize,
    interp_wcs,
)
from .abstract import AbstractPartialUniqSkymap
from .plot import plot, gridplot, _one_pt
from .plotters import (
    multiplot,
)
from .healpy import healpy as hp
from .points import PT_META_REGEX, PT_META_KW_REGEX, PT_META_COLOR_REGEX, _vecs_for_repr_, PointsTuple

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from astropy.visualization import wcsaxes
    from astropy.wcs.wcs import WCS
    import matplotlib.gridspec

# _depr_visufunc = deprecated(
#     reason=dedent("""
#         healpy visufunc plotting methods have been replaced with more powerful,
#         cross-platform WCSAxes-based plotters.  Use ``hpmoc.plot.plot`` and
#         ``PartialUniqSkymap.plot`` instead, for which this method is now a thin
#         wrapper. This method is retained for convenience, but its interface
#         has changed somewhat, and it may be removed or further modified without
#         warning in the future.
#     """.rstrip().strip('\n')).replace('\n', ' '),
#     version = "0.3.0"
# )

_depr_visufunc = lambda x: x

_XType = TypeVar('_XType', bound='np.generic')
_YType = TypeVar('_YType', bound='np.generic')
_OType = TypeVar('_OType', bound='np.generic')

@overload
def apply_diadic(
    op: Callable[[NDArray[_XType], NDArray[_YType]], NDArray[_OType]],
    x: 'PartialUniqSkymap[_XType]',
    y: 'PartialUniqSkymap[_YType]',
    /,
    **kwargs
) -> 'PartialUniqSkymap[_OType]': ...

@overload
def apply_diadic(
    op: Callable[[NDArray[_XType], Union[_YType, NDArray[_YType]]], NDArray[_OType]],
    x: 'PartialUniqSkymap[_XType]',
    y: Union[_XType, NDArray[_XType]],
    /,
    **kwargs
) -> 'PartialUniqSkymap[_OType]': ...

@overload
def apply_diadic(
    op: Callable[[Union[_XType, NDArray[_XType]], NDArray[_YType]], NDArray[_OType]],
    x: Union[_XType, NDArray[_XType]],
    y: 'PartialUniqSkymap[_YType]',
    /,
    **kwargs
) -> 'PartialUniqSkymap[_OType]': ...

def apply_diadic(f, x, y, /, iop=None, post=None, **kwargs):
    """
    Apply the diadic function ``f(x, y) -> o``. If both ``x`` and `y``, are 
    ``PartialUniqSkymap``s, the operation is performed efficiently on a MOC
    healpix grid either upsampling (if ``coarse=False``, default) to the
    highest input resolution in each area of the sky or downsampling to the
    lowest input resolution (if ``coarse=True``). Additional kwargs are passed
    to ``hpmoc.utils.uniq_diadic``. If one of ``x`` or ``y`` is not a skymap,
    the operation is the value array with standard numpy array/scalar arithmetic.

    The resulting ``PartialUniqSkymap`` will have the point sources from both
    ``x`` and ``y``. The metadata and history of ``x`` will be preserved,
    unless ``x`` is not a ``PartialUniqSkymap``.

    ``PartialUniqSkymap`` support all Python operators like *, +, -, and /
    already. This function should be used for calculations that require more
    than one basic arithmetic operation.

    Parameters
    ----------
    f : function (x, y) -> o
        The operation to apply
    x : PartialUniqSkymap or array-like
        The left side operand. At least one of ``x`` or ``y`` must be a
        ``PartialUniqSkymap``
    y : PartialUniqSkymap or array-like
        The right side operand. At least one of ``x`` or ``y`` must be a
        ``PartialUniqSkymap``
    iop : function (x, y) -> o, optional
        NUNIQ indices of the output skymap.
    **kwargs
        Additional kwargs are passed to ``hpmoc.utils.uniq_diadic``
    Returns
    -------
    o : PartialUniqSkymap
        The result of the operation ``f(x, y)``. If ``x`` and ``y`` are both
        ``PartialUniqSkymap``s, ``o`` may have different UNIQ indices from
        either due to resolution conformation.

    See Also
    --------
    hpmoc.utils.uniq_diadic

    """
    if iop is None:
        iop = f
    px = isinstance(x, PartialUniqSkymap)
    py = isinstance(y, PartialUniqSkymap)
    if px and py:
        uu, ss = uniq_diadic(f, (x.u, y.u), (x.s, y.s), **kwargs)
        pts = [*x.point_sources, *y.point_sources]
        xname = x.name or 'PIXELS'
        yname = y.name or 'PIXELS'
        s = x
    elif px:
        ss = f(x.s, y)
        uu = x.u
        pts = x.point_sources
        xname = x.name or 'PIXELS'
        yname = 'array'
        s = x
    elif py:
        ss = f(x, y.s)
        uu = y.u
        pts = y.point_sources
        xname = 'array'
        yname = y.name or 'PIXELS'
        s = y
    else:
        raise TypeError("At least one of `x` or `y` must be a `PartialUniqSkymap`.")

    meta = s.meta.copy()
    meta['HISTORY'] = meta.get('HISTORY', []) + [
        f'DIAD: {f.__name__}({xname}, {yname})']
    return PartialUniqSkymap(ss if post is None else post(ss), uu,
                                point_sources=pts, copy=False, meta=meta)
    

DIADIC_EXCEPTIONS = {'and': operator.and_, 'or': operator.or_,
                     'divmod': divmod}

def _get_op(name):
    if name in DIADIC_EXCEPTIONS:
        return DIADIC_EXCEPTIONS[name]
    return getattr(operator, name)


def diadic_dunder(pad=None, coarse: bool = False, post=None):
    """
    Implement diadic dunder methods like ``__add__``, ``__radd__``, etc. for
    scalar, array, and ``PartialUniqSkymap`` arguments using ``uniq_diadic``.
    ``pad`` and ``coarse`` are passed to ``uniq_diadic`` when
    performing operations between ``PartialUniqSkymap`` instances.
    If provided, run ``post`` on the result array before creating a new
    ``PartialUniqSkymap`` instance out of it (for example, to cast booleans to
    integers).
    """

    def decorator(meth):
        "The actual decorator."
        name = meth.__name__[2:-2]
        srt = lambda s, o: (s, o)
        if name.startswith('i'):
            Ω = Ωᵢ = _get_op(name)
            try:
                Ω = _get_op(name[1:])
            except AttributeError:
                pass
        elif name.startswith('r'):
            try:
                Ω = Ωᵢ = _get_op(name)
            except AttributeError:
                Ω = Ωᵢ = _get_op(name[1:])
                srt = lambda s, o: (o, s)
        else:
            Ω = Ωᵢ = _get_op(name)

        @functools.wraps(meth)
        def wrapper(s, o, pad=pad, coarse=coarse, post=post, **kwargs):
            try:
                return apply_diadic(Ω, *srt(s, o), iop=Ωᵢ, pad=pad, coarse=coarse, post=post, **kwargs)
            except TypeError:
                return NotImplemented

        wrapper.__doc__ = dedent(f"""
            ``__{name}__`` for scalars, arrays, and `PartialUniqSkymap`
            instances. Arrays must match ``s`` pixel-for-pixel. Provide keyword
            arguments ``pad`` to provide a pad value for missing pixels and/or
            ``coarse`` to specify whether the resulting skymap should take the
            higher or lower resolution in overlapping areas (default coarse
            value: {coarse}). For additional kwargs, see documentation of
            ``hpmoc.utils.uniq_diadic``.
        """)

        return wrapper

    return decorator


def bool_to_uint8(s):
    "Convert a boolean value to ``np.uint8``."
    import numpy as np

    return np.array(s, dtype=np.uint8)


_DType = TypeVar('_DType', covariant=True, bound='np.generic')
_Other_DType = TypeVar('_Other_DType', bound='np.generic')
class PartialUniqSkymap(AbstractPartialUniqSkymap[_DType]):
    """
    A HEALPix skymap object residing in memory with NUNIQ ordering. Only
    a subset of the full sky. You can index into a ``PartialUniqSkymap`` with
    NUNIQ indices to get a skymap with the same shape (optionally padding
    missing values with a second index argument). You can also use index
    notation to set pixel values at the specified NUNIQ index locations.
    """
    point_sources: List[PointsTuple]

    def __init__(
        self,
        s: 'NDArray[_DType]',
        u: Union['NDArray[np.integer[Any]]', 'WCS'],
        copy: bool = False, name=None, point_sources=None,
        meta=None, empty=None, compress=False, interp="nearest"):
        """
        Initialize a skymap with the pixel values and NUNIQ indices used.

        Parameters
        ----------
        s : array-like
            Pixel values. Must be numeric.
        u : array-like, WCS, or None
            NUNIQ indices corresponding to pixels in s. If ``None``, assume
            ``s`` is given as a single-resolution, all-sky NEST HEALPix skymap.
            Pass ``u='RING'`` to indicate a HEALPix RING-ordered all-sky
            skymap, or ``u='NEST'`` to explicitly indicate NEST ordering
            If a ``WCS`` instance, assume ``s`` is the image described by that
            ``WCS`` and convert the image into an MOC HEALPix image using
            bilinear interpolation (pixel sizes will locally be approximately
            the same as those in the original ``WCS``).
        copy : bool, optional
            Whether to make copies of the input arrays.
        name : str, optional
            The name of the skymap column. Used as the pixel column name when
            saving ``to_table``.
        point_sources : List[PointsTuple], optional
            If this skymap is associated with a list of point sources, you can
            provide it as an argument. These point sources will be included in
            data products for this skymap as well as plots.
        meta : OrderedDict, optional
            Metadata for this skymap. Used when saving to file. If this skymap
            was loaded from a file, this field will contain the metadata from
            that file. ``point_sources`` are removed from metadata before
            storing it in a ``PartialUniqSkymap``. ``PIXTYPE``, ``ORDERING``,
            and ``PARTIAL`` are set automatically.
        empty : scalar, optional
            Pixels with this value are interpreted as being empty and are
            discarded from the skymap on initialization to save storage space.
            This requires reindexing into the input arguments and therefore
            implies ``copy=True``. For example, set this to ``healpy.UNSEEN``
            to automatically discard pixels not included in a standard full-sky
            ``healpy`` skymap.
        interp : int, optional
            The interpolation strategy to use. Pass ``"nearest"`` for
            nearest-neighbor or ``"bilinear"`` for bilinear. See
            ``hpmoc.utils.interp_wcs`` for a full list of valid interpolation
            strategies. *Ignored if* ``u`` *is not a* ``WCS`` *instance.*

        Raises
        ------
        ValueError
            If ``s`` is not a numeric data type; if ``u`` is not provided and
            ``s`` cannot be interpreted as an all-sky fixed-resolution NEST
            HEALPix skymap; or if ``interp`` is not a valid value.
        """
        import numpy as np
        from astropy.wcs import WCS
        from astropy.units import Quantity as Qty
        from astropy.table.column import Column as Col

        if u is None or isinstance(u, str):
            idx = np.arange(len(s))
            ns = hp.npix2nside(len(s))
            if u is not None and u.upper() == 'RING':
                idx = hp.ring2nest(ns, idx)
            u = nest2uniq(idx, ns, in_place=True)
            del idx, ns
        elif isinstance(u, WCS):
            u, s = interp_wcs(u, s, interp=interp)
        if len(s) != len(u):
            raise ValueError(f"Must have same lengths: s={s}, u={u}")
        self.name = name
        if empty is None:
            self.s = np.array(s, copy=copy)
            self.u = np.array(u, copy=copy)
        else:
            si = s == empty
            self.s = np.array(s, copy=False)[si]
            self.u = np.array(u, copy=False)[si]
        check_valid_nuniq(self.u)
        if not np.issubdtype(self.s.dtype, np.number):
            raise ValueError(f"`s` must be numeric. got: {s}")

        # provide point sources and deduplicate
        self.point_sources = PointsTuple.dedup(*(point_sources or []))

        meta = meta or {}
        newmeta = OrderedDict()
        for k, v in meta.items():
            if not (PT_META_REGEX.match(k) or
                    PT_META_KW_REGEX.match(k) or
                    PT_META_COLOR_REGEX.match(k)):
                newmeta[k] = v
        newmeta['PIXTYPE'] = 'HEALPIX'
        newmeta['ORDERING'] = 'NUNIQ'
        newmeta['PARTIAL'] = True
        self.meta = newmeta

        if isinstance(s, (Qty, Col)):              # preserve astropy unit
            self.s = Qty(self.s, cast(Union[Qty, Col], s).unit, copy=False)

    @overload
    def nside(self, as_skymap: Literal[True], copy: bool = False, **kwargs) -> 'PartialUniqSkymap[np.integer[Any]]': ...

    @overload
    def nside(self, as_skymap: Literal[False] = False, copy: bool = False, **kwargs) -> 'NDArray[np.integer[Any]]': ...

    def nside(self, as_skymap=False, copy=False, **kwargs): # type: ignore
        """
        Pixel NSIDE values. If ``as_skymap=True``, return as a
        ``PartialUniqSkymap`` instance (with ``**kwargs`` passed to init).
        """
        import numpy as np

        n = uniq2nside(self.u)
        if as_skymap:
            u = np.array(self.u, copy=True) if copy else self.u
            m = self.meta.copy()
            m['HISTORY'] = m.get('HISTORY', []) + ['Take HEALPix NSIDE.']
            return type(self)(n, u, copy=False, name='NSIDE', meta=m,
                              point_sources=self.point_sources,
                              **kwargs)
        return n

    @overload
    def orders(self, as_skymap: Literal[True] = True, copy: bool = False, **kwargs) -> 'PartialUniqSkymap[np.integer[Any]]': ...

    @overload
    def orders(self, as_skymap: Literal[False], copy: bool = False, **kwargs) -> 'NDArray[np.integer]': ...

    def orders(self, as_skymap=False, copy=True, **kwargs): # type: ignore
        """
        HEALPix order values. If ``as_skymap=True``, return as a
        ``PartialUniqSkymap`` instance (with ``**kwargs`` passed to init).
        """
        import numpy as np

        o = uniq2order(self.u)
        if as_skymap:
            u = np.array(self.u, copy=True) if copy else self.u
            m = self.meta.copy()
            m['HISTORY'] = m.get('HISTORY', []) + ['Take HEALPix order.']
            return type(self)(o, u, copy=False, name='ORDER', meta=m,
                              point_sources=self.point_sources,
                              **kwargs)
        return o

    def copy(self: 'PartialUniqSkymap[_DType]') -> 'PartialUniqSkymap[_DType]':
        """
        Return a copy of this instance with separate copies of its data. The
        copy can be edited without affecting the original.
        """
        return type(self)(self.s, self.u, copy=True, name=self.name,
                          point_sources=self.point_sources, meta=self.meta)

    def astype(self, dtype: Type['_Other_DType'], copy=True, **kwargs) -> 'PartialUniqSkymap[_Other_DType]':
        """
        Return a new ``PartialUniqSkymap`` with the data-type of ``s`` set to
        ``dtype``. If ``copy=True``, always make sure both ``u`` and ``s`` are
        copies of the original data in the new array. Otherwise, re-use ``u``
        and (if possible given the provided ``dtype`` and ``**kwargs``) ``s``.
        ``copy`` and ``**kwargs`` are passed on to ``s.astype`` to make the
        conversion.
        """

        # https://github.com/microsoft/pyright/issues/5412
        return type(self)(self.s.astype(dtype, copy=copy, **kwargs), # type: ignore
                          self.u, name=self.name, meta=self.meta,
                          point_sources=self.point_sources)

    def to(self, *args, **kwargs):
        """
        Convert the units of this skymap's pixels (if they are stored as
        an ``astropy.units.Quantity`` instance).

        Parameters
        ----------
        *args, **kwargs
            Passed on to ``astropy.units.Quantity.to``.

        Raises
        ------
        TypeError
            If ``self.s`` is not a ``Quantity``.
        """
        from astropy.units import Quantity

        if not isinstance(self.s, Quantity):
            raise TypeError("Can only convert dimensions of a ``Quantity``")
        _s = cast(Quantity, self.s) # make pyright happy
        return type(self)(_s.to(*args, **kwargs), self.u,
                          copy=False, name=self.name,
                          meta=self.meta.copy(),
                          point_sources=self.point_sources)

    def compress(self, stype: Optional['np.dtype'] = None, utype: Optional['np.dtype'] = None, **kwargs) -> 'PartialUniqSkymap':
        """
        Eliminate redundant pixels with ``utils.uniq_minimize`` and store
        indices ``u`` in the smallest integer size that represents all values.

        Parameters
        ----------
        stype : type, optional
            If provided, store ``s`` as this type. Defaults to ``s.dtype``.
        utype : type, optional
            If provided, store ``u`` as this type. Defaults to the smallest
            ``int`` type required to store all values of ``u``.
        kwargs
            Keyword arguments to pass to ``hpmoc.utils.uniq_minimize``.

        Returns
        -------
        compressed : PartialUniqSkymap
            A compressed version of this skymap.

        See Also
        --------
        hpmoc.utils.uniq_minimize
        """
        u, s = uniq_minimize(self.u, self.s)
        if utype is None:
            utype = max_uint_type(self.u.max())
        s = s if stype is None else s.astype(stype)
        return type(self)(s, u.astype(utype), copy=False,
                          name=self.name, meta=self.meta.copy(),
                          point_sources=self.point_sources)

    def sort(self: 'PartialUniqSkymap[_DType]', copy=True) -> 'PartialUniqSkymap[_DType]':
        """
        Sort this skymap by UNIQ indices ``u`` (sorting ``s`` as well, of
        course). If ``copy=True``, copy ``u`` and ``s`` and return a new
        ``PartialUniqSkymap``; otherwise, sort them in-place and return this
        ``PartialUniqSkymap``.
        """
        ui = self.u.argsort()
        if not copy:
            self.u[:] = self.u[ui]
            self.s[:] = self.s[ui]
            return self
        return type(self)(self.s[ui], self.u[ui], name=self.name,
                          meta=self.meta, copy=True,
                          point_sources=self.point_sources)

    @property
    def value(self):
        """
        Get a dimensionless view of this skymap (no effect if ``s`` is not an
        ``astropy.units.Quantity``).
        """
        from astropy.units import Quantity

        if isinstance(self.s, Quantity):
            s = cast(Quantity, self.s).value # make pyright happy
            return type(self)(s, self.u, copy=False, name=self.name,
                            meta=self.meta,
                            point_sources=self.point_sources)
        return self

    def apply(
            self,
            func: Callable,
            copy: bool = True,
            quantity: bool = True
    ) -> 'PartialUniqSkymap':
        """
        Map a function to this skymap's pixels and return a new skymap whose
        values are the returned values of ``func``. Note that ``func(self.s)``
        must therefore return values corresponding to the same pixels as the
        input skymap. Use this to, for example, get the logarithm of this
        skymap as ``self.apply(np.log)``. If ``copy == True``, make a copy of
        the NUNIQ indices ``self.u``; otherwise, share them with the new
        skymap. To operate directly on the skymap value (even if it is stored
        as an ``astropy.units.Quantity``), use ``quantity = True``. Note that
        this will strip units from the result.
        """
        from astropy.units import Quantity

        m = self.meta.copy()
        mod = getattr(func, '__module__',
                      getattr(type(func), '__module__',
                              sys._getframe(-1).f_locals.get('__name__')))
        mod = '' if mod is None else mod + '.'
        m['HISTORY'] = (m.get('HISTORY', []) +
                        [f'Applied {mod}{func.__name__}'])
        n = f"{func.__name__}({self.name})"
        if not quantity and isinstance(self.s, Quantity):
            s = cast(Quantity, self.s).value
        else:
            s = self.s
        return type(self)(func(s), self.u.copy() if copy else self.u,
                          copy=False, name=n, meta=m,
                          point_sources=self.point_sources)

    def to_table(self, name=None, uname='UNIQ'):
        """
        Return a new ``astropy.table.Table`` whose ``UNIQ`` column is the NUNIQ
        indices ``self.u`` and ``PIXELS`` (or ``self.name``, if set) column is
        the skymap pixel values ``s``. Optionally override the pixel value
        column name and/or the NUNIQ column name with the ``name`` and
        ``uname`` arguments respectively.
        """
        from astropy.table import Table

        name = name or self.name or 'PIXELS'
        t = Table([self.u, self.s], names=[uname, name], meta=self.meta)
        for pt in self.point_sources:
            cast(MutableMapping, t.meta).update(PointsTuple(*pt).meta_dict())
        return t

    def write(
            self,
            file: Union[IO, str],
            *args,
            strategy: str = 'basic',
            **kwargs
    ):
        """
        Write a ``PartialUniqSkymap`` instance to file using the specified
        ``IoStrategy``. See ``IoRegistry`` attributes for details.

        Parameters
        ----------
        file : file or str
            The file object or filename to write to.
        *args
            Arguments to pass on to ``IoRegistry``.
        strategy : str, optional
            Name of the ``IoRegistry`` strategy to use for reads/writes.
        **kwargs
            Keyword arguments to pass on to a specific ``IoStrategy``. See
            ``IoRegistry`` properties for details.

        See Also
        --------
        astropy.table.Table.write
        astropy.table.Table.write.help
        astropy.io.fits.open
        IoRegistry
        IoRegistry.basic
        """
        from .io import IoRegistry

        return getattr(IoRegistry, strategy).write(self, file, *args, **kwargs)

    def read(
            *args,
            strategy: str = 'basic',
            **kwargs
    ):
        """
        Read a ``PartialUniqSkymap`` from file using
        ``astropy.table.Table.read``. See ``IoRegistry`` attributes for
        details. When called as a bound method or passed a
        ``PartialUniqSkymap`` as the first argument, uses that skymap as a
        mask for the bound skymap and only loads those pixels.

        Parameters
        ----------
        mask : PartialUniqSkymap, optional
            Only read pixels overlapping with this mask.
        file : file or str
            The file object or filename to read from.
        *args
            Arguments to pass on to ``IoRegistry``.
        strategy : str, optional
            Name of the ``IoRegistry`` strategy to use for reads/writes.
        **kwargs
            Keyword arguments to pass on to a specific ``IoStrategy``. See
            ``IoRegistry`` properties for details.

        Returns
        -------
        m : PartialUniqSkymap
            A new ``PartialUniqSkymap`` instance with the specified data.

        Examples
        --------
        For multi-skymap fits files, you can load the full set of skymaps with
        ``hdul = astropy.io.fits.open(fitsfile)`` and then load the skymap of
        interest with ``PartialUniqSkymap.read(hdul[i])``. With fits files, you
        can memory-map the resulting skymap to save memory (at the expense of
        speed) by adding ``memmap=True`` as a keyword argument.

        See Also
        --------
        astropy.table.Table.read
        astropy.table.Table.read.help
        astropy.io.fits.open
        IoRegistry
        IoRegistry.basic
        """
        from .io import IoRegistry

        if not isinstance(args[0], PartialUniqSkymap):  # unbound method; add
            args = (None,) + args                       #   placeholder to args
        return getattr(IoRegistry, strategy).read(*args, **kwargs)

    def fill(self, nside=None, pad=None, as_skymap=False):
        """
        Return a full-sky *nested* HEALPix skymap at NSIDE resolution
        ``nside``.

        Parameters
        ----------
        nside : int
            HEALPix NSIDE value of the output map. If not provided, use the
            highest NSIDE value in this skymap's ``nside`` values to preserve
            detail.
        pad : float, optional
            Fill in missing values with ``pad`` (if not provided, use
            ``healpy.UNSEEN``). Preserves ``astropy.units.Unit`` of this
            skymap's pixel values (if ``s`` is an ``astropy.units.Quantity``).
        as_skymap : bool, optional
            If ``True``, return a ``PartialUniqSkymap`` instance with the new
            pixelization (instead of a bare array with implicit indexing).

        Returns
        -------
        s : array or PartialUniqSkymap
            The filled-in skymap, either as an array if ``as_skymap == False``
            or as a new ``PartialUniqSkymap`` instance.

        See Also
        --------
        PartialUniqSkymap.fixed
        """
        import numpy as np

        nside = nside or uniq2nside(self.u.max())
        sᵒ = fill(self.u, self.s, nside, pad=pad)
        if not as_skymap:
            return sᵒ
        m = self.meta.copy()
        m['HISTORY'] = (m.get('HISTORY', []) +
                        [f'Filled to NEST, NSIDE={nside}.'])
        return PartialUniqSkymap(sᵒ, np.arange(4*nside**2, 16*nside**2),
                                 copy=False, meta=m,
                                 point_sources=self.point_sources)

    def quantiles(
            self: 'PartialUniqSkymap[_DType]',
            quantiles: Sequence[float],
    ) -> Tuple[Iterator, 'NDArray[_DType]', '_DType']:
        """
        Get an iterator of downselected skymaps partitioned by ``quantiles``.
        For example, get the smallest sky area containing 90% of the
        probability (or whatever other intensive quantity this skymap
        represents) with ``quantiles=[0.1, 1]``.

        Parameters
        ----------
        quantiles : Sequence[float]
            Quantiles from which to select pixels. Must be in ascending order
            with values in the interval ``[0, 1]``. These will form endpoints
            for partitions of the ``skymap``. For example, ``[0.1, 0.9]`` will
            omit the lowest and highest value pixels, giving the intermediate
            pixels accounting for 80% of the integrated skymap.  Note that
            quantiles returned by this function are non-intersecting and
            half-open on the right (as with python indices), with the exception
            of ``1`` for the last pixel; for example, ``[0, 1]`` will include
            all pixels, ``[0.5, 1]`` will include the highest density pixels
            accounting for 50% of the integrated skymap value, ``[0, 0.5, 1]``
            will partition the skymap into non-intersecting sets of pixels
            accounting for the high- and low-density partitions of the skymap
            by integrated value, etc.

        Returns
        -------
        partitions: Iterator[PartialUniqSkymap]
            A generator containing non-overlapping skymaps falling into the
            specified ``quantiles``. Will have length one less than the number
            of quantiles, which form the boundaries of the partitions. Always
            an iterable, even if only two quantiles are provided; you can
            unpack a single value with, e.g., ``x, = m.quantiles(...)``.
        levels: astropy.units.Quantity or array
            Values of the skymap at each quantile. Useful for e.g. contour
            plots (though ``PartialUniqSkymap.plot`` will handle this
            automatically).
        norm: astropy.units.Quantity or float
            The integrated value of the skymap. Useful for calculating the
            integral of each quantile.

        Raises
        ------
        ValueError
            If ``quantiles`` has length less than 2; if its values are not in
            order and contained in the interval ``[0, 1]``; if ``nside`` and
            ``skymap`` cannot be broadcast together; if any values in
            ``skymap`` are negative; or if the total integrated skymap equals
            zero, in which case quantiles are undefined.

        See Also
        --------
        hpmoc.utils.nside_quantile_indices
        """
        import numpy as np

        quantiles_arr = np.asarray(quantiles, dtype=np.float_)
        indices, levels, norm = nside_quantile_indices(self.nside(), self.s,
                                                       quantiles_arr)

        def skymaps():
            for i, l, u in zip(indices, quantiles_arr[:-1], quantiles_arr[1:]):
                m = self.meta.copy()
                m['HISTORY'] = m.get('HISTORY', []) + wrap(
                    f'Downselected to [{l:.2g}, {u:.2g}] quantile '
                    f'({(u-l)*100:.2g}%) of {norm:.2g} ({norm*(u-l):.2g} ',
                    70
                )
                yield PartialUniqSkymap(self.s[i], self.u[i], copy=False,
                                        meta=m,
                                        point_sources=self.point_sources)

        return skymaps(), levels, norm

    def fixed(self, nside=None):
        """
        Re-raster to a fixed NSIDE. Like ``fill`` but for partial skymaps.

        Parameters
        ----------
        nside : int
            HEALPix NSIDE value of the output map. If not provided, use the
            highest NSIDE value in this skymap's ``nside`` values to preserve
            detail.
        """
        nside = nside or uniq2nside(self.u.max())
        u⃗ᵒ = uniq2nest(self.u, nside, nest=False)
        s = self.reraster(u⃗ᵒ, copy=False)
        s.meta['HISTORY'][-1] += f' (fixed NSIDE={nside})'
        return s

    def __getitem__(self, idx):
        """
        Get a view into this skymap with the given index applied to ``u`` and
        ``s``. Uses their provided ``__getitem__`` semantics, so you'll get
        e.g. a view on the same data if using a slice index.

        Note that the return value will *always* be a ``PartialUniqSkymap``,
        even if you provide a scalar index; scalar return values are made into
        length-1 lists, as if you'd asked for a slice of length 1.

        **NB: the provided indices are treated as simple array indices, NOT as
        UNIQ indices; order matters!**

        If you want to guarantee a copy on a view that you know is not copied,
        make a copy with the returned array.
        """
        import numpy as np

        m = self.meta.copy()
        repidx = repr(idx)
        msg = repidx if len(repidx) < 60 else repidx[:58]+'...'
        m['HISTORY'] = m.get('HISTORY', []) + [f'Got view: {msg}']
        s = np.atleast_1d(self.s[idx])
        u = np.atleast_1d(self.u[idx])
        return type(self)(s, u, copy=False, meta=m,
                          point_sources=self.point_sources)

    def __setitem__(self, idx, value):
        """
        Like ``__getitem__``, will set values using the semantics of ``s``
        datatype.
        """
        raise NotImplementedError("Revisit this.")
        self[idx].s = value

    def _iparser(self, item):
        """
        Parse the ``item`` argument for ``__getitem__``, ``__setitem__``, and
        ``__delitem__``.
        """

    def intersection(self, u):
        """
        See ``utils.uniq_intersection``.
        """
        if isinstance(u, AbstractPartialUniqSkymap):
            u = u.u
        return uniq_intersection(self.u, u)

    def render(self, u⃗ᵒ, pad=None, valid=None, mask_missing=False):
        """
        Like ``reraster``, but ``u⃗ᵒ`` does not need to be unique. Use this to
        e.g. render a skymap to a plot. Unlike ``reraster``, will not return a
        ``PartialUniqSkymap``; instead, simply returns the pixel values
        corresponding to ``u⃗ᵒ``.

        ``u⃗ᵒ`` can also be an ``astropy.wcs.WCS`` world coordinate system, in
        which case the returned array will contain the pixel values of this
        skymap in that coordinate system (with regions outside of the
        projection set to ``np.nan``).

        Parameters
        ----------
        u⃗ᵒ: array or astropy.wcs.WCS
            The pixels to fill. If an array, treated as UNIQ indices
            (duplicates allowed); if WCS, treated as a set of pixels to render
            to.
        pad: float, optional
            A pad value to use for pixels not contained in the maps. Defaults
            to ``None``, which will raise an error if suitable values cannot
            be found for every valid pixel in ``u⃗ᵒ`` (this does not apply to
            values outside a ``WCS`` projection, which will take on ``np.nan``
            values).
        valid: array, optional
            If provided, results will be scattered into an array of the same
            shape as ``valid``, filling the indices where ``valid==True``. The
            number of ``True`` values in ``valid`` must therefore equal the
            length of ``u⃗ᵒ``. This argument only makes sense if ``u⃗ᵒ`` is an
            array of NUNIQ indices; if it is a ``WCS`` instance and ``valid``
            is provided, an error is raised. Use ``valid`` to produce plots or
            to reuse indices produced by ``wcs2mask_and_uniq`` in several
            ``render`` invocations.  See note on how ``mask_missing`` affects
            the result.
        mask_missing : bool
            If ``mask_missing=True``, return a ``np.ma.core.MaskedArray``.
            Missing values are tolerated and are marked as ``True`` in the
            ``mask_missing``. They will be set to ``pad or 0`` in the ``data``
            field. If ``valid`` is also provided, then the output will still be
            a ``np.ma.core.MaskedArray``, but will be set to ``True`` wherever
            ``valid == False`` in addition to wherever pixels are missing (and
            will still take on masked values of ``np.nan`` in the invalid
            regions).

        Returns
        -------
        pixels: array
            Pixel values at locations specified in ``u⃗ᵒ``. If ``u⃗ᵒ`` is a
            ``WCS`` instance, then values outside of the projection will be
            set to ``np.nan``.
        """
        return render(self.u, self.s, u⃗ᵒ, pad=pad, valid=valid,
                      mask_missing=mask_missing)

    def reraster(self, u_out, pad=None, mask_missing=False, copy=True):
        """
        Return a new ``PartialUniqSkymap`` instance with the same pixel values
        rerasterized to match the output NUNIQ indices ``u_out``. Fill in missing
        values in the output skymap with ``pad``. If ``pad`` is not provided
        and this skymap does not cover the full region defined in ``u_out``,
        raises a ``ValueError``. Preserves ``astropy.units.Unit`` of this
        skymap's pixel values (if ``s`` is an ``astropy.units.Quantity``). If
        ``copy`` is ``False``, use ``u_out`` as the indices of the new skymap;
        otherwise, use a copy.
        """
        import numpy as np

        s_out = reraster(self.u, self.s, u_out, pad=pad, mask_missing=mask_missing)
        m = self.meta.copy()
        m['HISTORY'] = m.get('HISTORY', []) + ['Rerasterized.']
        return PartialUniqSkymap(s_out, np.array(u_out, copy=copy), copy=False,
                                 meta=m, point_sources=self.point_sources)

    def coords(self):
        """
        Get the sky coordinates (right-ascension and declination, ICRS)
        corresponding to each pixel in the skymap.

        Returns
        -------
        ra_dec : astropy.units.Quantity
            2D array whose first row is the right-ascension and second row is
            the declination (in degrees) of each pixel. You can get each of
            these individually with ``ra, dec = self.coords()``.
            ``self.coords()[:, i]`` corresponds to RA, Dec for ``self.s[i]``.
        """
        return nest2ang(*uniq2nest_and_nside(self.u))

    def area(self):
        "Area per-pixel for pixels in this skymap in ``astropy.unit.sr``."
        from astropy.units import sr  # pylint: disable=no-name-in-module
        return nside2pixarea(self.nside(), degrees=False)*sr     # steradian

    def ang_dist(self, ra, dec, degrees=True):
        """
        Get distances from each pixel in this skymap to the point at
        right-ascension ``ra`` and declination ``dec``.

        Parameters
        ----------
        ra : array-like or astropy.units.Quantity
            Right-ascension of the point
        dec : array-like or astropy.units.Quantity
            Declination of the point
        degrees : bool, optional
            If ``ra`` and ``dec`` are ``astropy.units.Quantity`` instances,
            they will be automatically converted. If they are unitless scalars,
            they will be interpreted as degrees if ``degrees=True``, radians
            otherwise.

        Returns
        -------
        ang_dist : astropy.units.Quantity
            The distances of each pixel in this skymap to the point at ``ra``,
            ``dec`` in degrees.

        Examples
        --------
        We should find that the distance from any pixel to the North pole is
        equal to 90 minus the declination (within some small error):

        >>> import numpy as np
        >>> from astropy.units import deg
        >>> skymap = PartialUniqSkymap(*([4+np.arange(12)]*2))
        >>> _, dec = skymap.coords()
        >>> dec
        <Quantity [ 41.8103149,  41.8103149,  41.8103149,  41.8103149,   0.       ,
                     0.       ,   0.       ,   0.       , -41.8103149, -41.8103149,
                   -41.8103149, -41.8103149] deg>
        >>> Δθ⃗ = skymap.ang_dist(32, 90)
        >>> Δθ⃗
        <Quantity [0.84106867, 0.84106867, 0.84106867, 0.84106867, 1.57079633,
                   1.57079633, 1.57079633, 1.57079633, 2.30052398, 2.30052398,
                   2.30052398, 2.30052398] rad>
        >>> np.all(abs(Δθ⃗+dec-90*deg).to('deg').value<1e-13)
        True

        Likewise, the distance from any pixel to the South pole should be
        equal to 90 plus the declination:

        >>> not np.around(skymap.ang_dist(359, -90)-dec-90*deg,
        ...               14).value.any()
        True
        """
        return uniq2dangle(self.u, ra, dec, degrees=degrees)

    def unzip_orders(self):
        """
        Return a list of ``PartialUniqSkymap`` instances corresponding to the
        parts of this sky imaged and each HEALPix order. Length equals the
        maximum order of this skymap. Empty terms indicate that this skymap
        does not have pixels of the corresponding HEALPix order.
        """
        # TODO test
        srt = self.sort()
        [[s], o⃗] = nside_slices(srt.u)[1:3]
        return [srt[s] for s in [slice(0, 0)]*o⃗[0]+s]

    def unzip_atlas(self):
        "Return 12 sub-skymaps corresponding to the HEALPix base pixels."
        raise NotImplementedError()

    def min(self):
        "Minimum skymap value = ``self.s.min()``."
        return self.s.min()

    def max(self):
        "Maximum skymap value = ``self.s.max()``."
        return self.s.max()

    @property
    def unit(self):
        "``self.s.unit``, if defined; otherwise ``None``."
        return getattr(self.s, 'unit', None)

    @_depr_visufunc
    def azeqview(self, *scatter, **kwargs):
        return self.plot(*scatter, projection='ARC', **kwargs)

    @_depr_visufunc
    def cartview(self, *scatter, **kwargs):
        return self.plot(*scatter, projection='CAR', **kwargs)

    @_depr_visufunc
    def gnomview(self, *scatter, **kwargs):
        return self.plot(*scatter, projection='TAN', **kwargs)

    @_depr_visufunc
    def orthview(self, *scatter, **kwargs):
        return self.plot(*scatter, projection='SIN', **kwargs)

    @_depr_visufunc
    def mollview(self, *scatter, **kwargs):
        return self.plot(*scatter, projection='MOL', **kwargs)

    def plot(self, *scatter: PointsTuple, **kwargs) -> Union[
            'wcsaxes.WCSAxes',
            'wcsaxes.WCSAxesSubplot'
    ]:
        """
        Plot this skymap. A thin wrapper around ``hpmoc.plot.plot`` that
        automatically includes scatter plots for this instance's
        ``point_sources``.

        Parameters
        ----------
        scatter: PointsTuple
            Additional point sources to plot (in addition to those stored
            in ``self.point_sources``).
        kwargs
            Keywork arguments to be passed to ``hpmoc.plot.plot``.

        Returns
        -------
        ax: WCSAxes or WCSAxesSubplot
            The axes that were just plotted to.

        See Also
        --------
        hpmoc.plot.plot
        """
        return plot(self, *(*self.point_sources, *scatter), **kwargs)

    def gridplot(
            self,
            *skymaps: Union[
                'PartialUniqSkymap',
                'NDArray[Any]',
                Tuple[
                    'NDArray[Any]',
                    Optional[
                        Union[
                            'NDArray[Any]',
                            'WCS',
                            str,
                        ]
                    ],
                ],
            ],
            **kwargs
    ) -> Tuple[
            'matplotlib.gridspec.GridSpec',
            Sequence[Sequence['wcsaxes.WCSAxes']]
    ]:
        """
        Plot this skymap and any others in ``skymaps``. A thin wrapper around
        ``hpmoc.plot.gridplot``.

        Parameters
        ----------
        *skymaps : PartialUniqSkymap or map-like
            Skymaps to pass to ``gridplot``. Can be anything accepted by
            ``hpmoc.plot.plot``, which ``hpmoc.plot.gridplot`` will use to
            display them.
        **kwargs
            Keyword arguments for ``hpmoc.plot.gridplot``.

        See Also
        --------
        hpmoc.plot.gridplot
        """
        return gridplot(self, *skymaps, **kwargs)

    @_depr_visufunc
    def multiplot(self, *skymaps: Sequence['PartialUniqSkymap'], nest: bool = True, **kwargs):
        """
        Call ``plotters.multiplot`` with the default ``transform``
        suitable for a ``PartialUniqSkymap``.

        Parameters
        ----------
        *skymapsₗ : List[Union[PartialUniqSkymap, array]]
            Skymaps to plot. Can be ``PartialUniqSkymap`` instances or
            full-sky single-resolution skymaps.
        **kwargs
            Keyword arguments passed to ``plotters.multiplot``.

        Returns
        -------
        fig : matplotlib.figure.Figure
            A new ``matplotlib`` figure containing the specified subplots.

        See Also
        --------
        plot.multiplot
        PartialUniqSkymap.plot
        plot.plot
        """
        return multiplot(self, *skymaps, **kwargs)

    def _vecs_for_repr_(self, maxlen, *vecs):
        if not vecs:
            vecs = self.u, self.s
        return _vecs_for_repr_(maxlen, *vecs)

    def __str__(self):
        return self.to_table().__str__()

    def __repr__(self):
        return self.to_table().__repr__()

    def _repr_html_(self):
        [u, s], [_, unit] = self._vecs_for_repr_(20)
        pts = self.point_sources
        unit = f'<thead><tr><th></th><th>{unit}</th></tr></thead>'
        rows = "\n".join(f'<tr><td>{u}</td><td>{s}</td></tr>'
                         for u, s in zip(u, s))
        meta_chunks = [
            (
                k,
                type(v).__name__,
                (
                    f"<td>{v}</td>" if not isinstance(v, (list, tuple)) else
                    "<td><table>{}</table></td>".format(
                        "\n".join(f"<tr><td>{vv}</td></tr>" for vv in v))
                )
            ) for k, v in  self.meta.items()
        ]
        meta = "\n".join(f'<tr><th>{k}</th><td><em>{t}</em></td>{c}</tr>'
                         for k, t, c in meta_chunks)
        name = self.name or 'PIXELS'
        pt_srcs = "\n".join(f'<div><h5>Point Sources {i}</h5>'
                            f'{PointsTuple(*p)._repr_html_()}</div>'
                            for i, p in enumerate(self.point_sources))

        if 'matplotlib' in sys.modules:
            import matplotlib.pyplot as plt

            img = BytesIO()
            projs = ['CEA']
            widths = [2]
            one_pt = _one_pt(pts, None)
            if one_pt:
                projs.append('ARC')
                widths.append(1)
            kwargs = {'projections': projs, 'bottom': 0.1, 'left': 0.04,
                      'missing_color': 'gray'}
            if 'IPython' in sys.modules:
                from IPython.utils.capture import capture_output

                with capture_output():
                    gs, [[ax], *_] = self.gridplot(**kwargs)
            else:
                gs, [[ax], *_] = self.gridplot(**kwargs)
            if one_pt:
                for co in ax.coords:
                    co.set_ticklabel_visible(False)
            fig = gs.figure
            assert fig is not None
            fig.savefig(img, format='png')
            plt.close(fig)
            img.seek(0)
            #b64 = "".join(base64.encodebytes(img.read()).decode().split('\n'))
            b64 = base64.b64encode(img.read()).decode()
            pd=f'''
                <div style="vertical-align: top;">
                    <h5>Plot</h5>
                    <img src="data:image/png;base64,{b64}" alt="Preview plot"
                         style="min-width: 200px; max-width: 800px;"/>
                </div>
            '''
        else:
            pd = '''
                <div>
                    <em>Import matplotlib to display plot preview.</em>
                </div>
            '''

        tab = f'''
        <style>
            .partialuniq_flexbox {{
                display: flex;
                flex-wrap: wrap;
                pad: -0.5em;
            }}

            .partialuniq_flexbox > div {{
                margin: 0.5em;
            }}

            .partialuniq_flexbox_vert {{
                flex-direction: column;
            }}
        </style>
        <div class="partialuniq_flexbox">
                <div class="partialuniq_flexbox partialuniq_flexbox_vert">
                    {pd}
                    {pt_srcs}
                </div>
                <div style="vertical-align: top;">
                    <h5>Skymap ({len(self.s)} pixels)</h5>
                    <table>
                        <thead><tr><th>UNIQ</th><th>{name}</th></tr></thead>
                        {unit}
                        <thead><tr><th>int64</th><th>float64</th></tr></thead>
                        {rows}
                    </table>
                </div>
                <div style="vertical-align: top;">
                    <h5>Metadata</h5>
                    <table>
                        {meta}
                    </table>
                </div>
        </div>
        '''
        return tab

    # BEGIN COMPARATOR METHODS
    # None of these methods should ever be reached, since they will be replaced
    # by a constructed method that calls ``uniq_diadic`` with the appropriate
    # operator. See ``diadic_dunder`` for more information.  

    @diadic_dunder(post=bool_to_uint8)
    def __eq__(self, other): raise NotImplementedError("__eq__")

    @diadic_dunder(post=bool_to_uint8)
    def __ne__(self, other): raise NotImplementedError("__ne__")

    @diadic_dunder(post=bool_to_uint8)
    def __le__(self, other): raise NotImplementedError("__le__")

    @diadic_dunder(post=bool_to_uint8)
    def __lt__(self, other): raise NotImplementedError("__lt__")

    @diadic_dunder(post=bool_to_uint8)
    def __ge__(self, other): raise NotImplementedError("__ge__")

    @diadic_dunder(post=bool_to_uint8)
    def __gt__(self, other): raise NotImplementedError("__gt__")

    # BEGIN NUMERIC METHODS

    @diadic_dunder()
    def __add__(self, other): raise NotImplementedError("__add__")

    @diadic_dunder()
    def __sub__(self, other): raise NotImplementedError("__sub__")

    @diadic_dunder()
    def __mul__(self, other): raise NotImplementedError("__mul__")

    # @diadic_dunder()
    # def __matmul__(self, other): raise NotImplementedError("__matmul__")

    @diadic_dunder()
    def __truediv__(self, other): raise NotImplementedError("__truediv__")

    @diadic_dunder()
    def __floordiv__(self, other): raise NotImplementedError("__floordiv__")

    @diadic_dunder()
    def __mod__(self, other): raise NotImplementedError("__mod__")

    @diadic_dunder()
    def __divmod__(self, other): raise NotImplementedError("__divmod__")

    @diadic_dunder()
    def __pow__(self, other): raise NotImplementedError("__pow__")

    @diadic_dunder()
    def __lshift__(self, other): raise NotImplementedError("__lshift__")

    @diadic_dunder()
    def __rshift__(self, other): raise NotImplementedError("__rshift__")

    @diadic_dunder()
    def __and__(self, other): raise NotImplementedError("__and__")

    @diadic_dunder()
    def __xor__(self, other): raise NotImplementedError("__xor__")

    @diadic_dunder()
    def __or__(self, other): raise NotImplementedError("__or__")

    # REVERSE NUMERIC METHODS

    @diadic_dunder()
    def __radd__(self, other): raise NotImplementedError("__radd__")

    @diadic_dunder()
    def __rsub__(self, other): raise NotImplementedError("__rsub__")

    @diadic_dunder()
    def __rmul__(self, other): raise NotImplementedError("__rmul__")

    # @diadic_dunder()
    # def __rmatmul__(self, other): raise NotImplementedError("__rmatmul__")

    @diadic_dunder()
    def __rtruediv__(self, other): raise NotImplementedError("__rtruediv__")

    @diadic_dunder()
    def __rfloordiv__(self, other): raise NotImplementedError("__rfloordiv__")

    @diadic_dunder()
    def __rmod__(self, other): raise NotImplementedError("__rmod__")

    @diadic_dunder()
    def __rdivmod__(self, other): raise NotImplementedError("__rdivmod__")

    @diadic_dunder()
    def __rpow__(self, other): raise NotImplementedError("__rpow__")

    @diadic_dunder()
    def __rlshift__(self, other): raise NotImplementedError("__rlshift__")

    @diadic_dunder()
    def __rrshift__(self, other): raise NotImplementedError("__rrshift__")

    @diadic_dunder()
    def __rand__(self, other): raise NotImplementedError("__rand__")

    @diadic_dunder()
    def __rxor__(self, other): raise NotImplementedError("__rxor__")

    @diadic_dunder()
    def __ror__(self, other): raise NotImplementedError("__ror__")

    # IN-PLACE NUMERIC METHODS

    @diadic_dunder()
    def __iadd__(self, other): raise NotImplementedError("__iadd__")

    @diadic_dunder()
    def __isub__(self, other): raise NotImplementedError("__isub__")

    @diadic_dunder()
    def __imul__(self, other): raise NotImplementedError("__imul__")

    # @diadic_dunder()
    # def __imatmul__(self, other): raise NotImplementedError("__imatmul__")

    @diadic_dunder()
    def __itruediv__(self, other): raise NotImplementedError("__itruediv__")

    @diadic_dunder()
    def __ifloordiv__(self, other): raise NotImplementedError("__ifloordiv__")

    @diadic_dunder()
    def __imod__(self, other): raise NotImplementedError("__imod__")

    @diadic_dunder()
    def __ipow__(self, other): raise NotImplementedError("__ipow__")

    @diadic_dunder()
    def __ilshift__(self, other): raise NotImplementedError("__ilshift__")

    @diadic_dunder()
    def __irshift__(self, other): raise NotImplementedError("__irshift__")

    @diadic_dunder()
    def __iand__(self, other): raise NotImplementedError("__iand__")

    @diadic_dunder()
    def __ixor__(self, other): raise NotImplementedError("__ixor__")

    @diadic_dunder()
    def __ior__(self, other): raise NotImplementedError("__ior__")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import hpmoc

# -- Project information -----------------------------------------------------

project = 'HEALPix Multi-Order Coordinate Library (HPMOC)'
copyright = '2021, Stefan Trklja Countryman'
author = 'Stefan Trklja Countryman'

# The short X.Y version
version = ".".join(hpmoc.__version__.split(".")[:2])
# The full version, including alpha/beta/rc tags
release = hpmoc.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'nbsphinx',  # include jupyter notebooks
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.graphviz',
    # 'sphinx.ext.mathjax', # excluded in LLAMA, likely performance issues?
    'sphinx.ext.imgmath',
    # 'sphinx.ext.viewcode', # do not want to include source in docs
    #'sphinx.ext.githubpages',  # add .nojekyll and CNAME
    'sphinx.ext.napoleon',  # understand NumPy and Google docstring syntax
    #'sphinxarg.ext',  # understand CLIs https://sphinx-argparse.readthedocs.io; conda install sphinx-argparse
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}
html_theme_options = {
    'canonical_url': 'http://hpmoc.stc.sh',
    # 'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    # 'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': True,
    # 'style_nav_header_background': 'black',
    # Toc options
    'collapse_navigation': False,
    'sticky_navigation': True,
    'navigation_depth': 5,
    # 'includehidden': True,
    # 'titles_only': False
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for LaTeX output ------------------------------------------------

latex_show_urls = 'footnote'
latex_engine = 'xelatex'

# preamble example taken from
# https://sphinxguide.readthedocs.io/en/latest/sphinx_basics/settings.html
latex_elements = {
    'printindex': '\\footnotesize\\raggedright\\printindex',
    'preamble': r'''
        \usepackage[utf8]{inputenc}
        \usepackage{amssymb}
        \usepackage{newunicodechar}

        \newunicodechar{ℝ}{$\mathbb{R}$}
        \newunicodechar{ρ}{$\rho$}
        \newunicodechar{Δ}{$\Delta$}
        \newunicodechar{δ}{$\delta$}
        \newunicodechar{Ω}{$\Omega$}
        \newunicodechar{ω}{$\omega$}
        \newunicodechar{Θ}{$\Theta$}
        \newunicodechar{θ}{$\theta$}
        \newunicodechar{ϵ}{$\epsilon$}
        \newunicodechar{ν}{$\nu$}
        \newunicodechar{ₙ}{$_{n}$}
        \newunicodechar{ₕ}{$_{h}$}
        \newunicodechar{ᶜ}{$^{c}$}
        \newunicodechar{ˢ}{$^{s}$}
        \newunicodechar{ᵉ}{$^{e}$}
        \newunicodechar{ᵐ}{$^{m}$}
        \newunicodechar{ᵖ}{$^{p}$}
    ''',

    # ADD UNICODE SUPPORT
    # 'utf8extra': r'''
    #     \ifdefined\DeclareUnicodeCharacter
    #         \ifdefined\DeclareUnicodeCharacterAsOptional
    #         \else
    #             \DeclareUnicodeCharacter{00A0}{\nobreakspace}\
    #             \DeclareUnicodeCharacter{3C1}{\nobreakspace}\
    #         \fi
    #     \fi
    # '''
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'llama.tex', 'Low-Latency Algorithm for Multi-messenger Astrophysics (LLAMA)',
     'Stefan Trklja Countryman', 'manual'),
]

# -- Extension configuration -------------------------------------------------

# add graphviz directives to each filehandler class found in llama
graphviz_output_format = 'svg'  # use SVG images for webpage output




# Copyright (C) 2022-2023 Columbia Experimental Gravity Group (GECo)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Strategies for loading LIGO/Virgo/KAGRA skymaps into ``PartialUniqSkymap``
instances.
"""

from typing import Optional, Union, IO, TYPE_CHECKING
from hpmoc.io.abstract import IoStrategy
from hpmoc.io.fits import load_ligo
from hpmoc.partial import PartialUniqSkymap
from hpmoc.utils import uniq_coarsen, uniq_minimize

import numpy as np
from numpy.typing import ArrayLike

class LigoIo(IoStrategy):
    """
    Read/write files in the format used by LIGO/Virgo for their skymaps.
    """

    @classmethod
    def read(
        cls,
        skymap: Optional[Union[PartialUniqSkymap, ArrayLike]],
        file: Union[IO, str],
        *args,
        name: Optional[str] = None,
        coarsen: Optional[int] = None,
        **kwargs
    ):
        """
        Read a file saved in the format used by LIGO/Virgo for their skymaps.

        Parameters
        ----------
        mask : PartialUniqSkymap or array, optional
            Only read in pixels overlapping with ``mask``.
        file : file or str
            The file object or filename to read from. Can be a stream as no
            seeking will be performed.
        name : str, optional
            The column-name of the pixel data. Defaults to 'PROBDENSITY'
        coarsen : int, optional
            If provided, coarsen the ``mask`` by up to this many HEALPix
            orders (up to order 0) to speed up read times. This will select
            a superset of the sky region defined in ``mask``.
        *args, **kwargs
            Arguments to pass on to ``hpmoc.fits.load_ligo``.
        """
        if name is None:
            name = 'PROBDENSITY'

        pt = skymap.point_sources if isinstance(skymap, PartialUniqSkymap) else []
        if skymap is not None:
            mask = skymap.u if isinstance(skymap, PartialUniqSkymap) else skymap
            mask = np.asarray(mask, dtype=np.int64)
            mask = uniq_coarsen(mask, coarsen) if coarsen is not None else mask
            mask = uniq_minimize(mask)[0]
        else:
            mask = None
        [[u, s, meta]] = load_ligo(file, mask=mask, **kwargs)
        return PartialUniqSkymap(s, u, name=name, meta=meta, point_sources=pt)

    @classmethod
    def write(
        cls,
        skymap: PartialUniqSkymap,
        file: Union[IO, str],
        name: Optional[str] = None,
        *args,
        **kwargs
    ):
        """
        Write a skymap to file in the format used by LIGO/Virgo for their
        skymaps. A thin wrapper around ``BasicIo.write``.
        """
        from hpmoc.io import BasicIo

        BasicIo.write(skymap, file, name=name, *args, **kwargs)



# Copyright (C) 2019-2023 Columbia Experimental Gravity Group (GECo)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"HEALPix helper utilities needed for both ``utils`` and ``healpy``."

import numpy as np

COMPRESS_MASKS = np.array([
    0x5555555555555555,
    0x3333333333333333,
    0x0f0f0f0f0f0f0f0f,
    0x00ff00ff00ff00ff,
    0x0000ffff0000ffff,
    0x00000000ffffffff,
], dtype=np.uint64)


def alt_compress(x, in_place=False):
    """
    Start in 0x55... state.

    https://help.dyalog.com/18.0/Content/Language/Primitive%20Functions/Replicate.htm

    Examples
    --------
    >>> alt_compress(0b011101)
    7
    >>> alt_compress(0b110010)
    4
    >>> alt_compress(100)
    10
    >>> f'{alt_compress(0b10011100):04b}'
    '0110'

    See Also
    --------
    alt_expand
    """

    if isinstance(x, int):
        x = np.uint64(x)
    elif not (in_place and np.issubdtype(x.dtype, np.uint64)):
        x = x.astype(np.uint64)
    x &= COMPRESS_MASKS[0]
    for i, m in enumerate(COMPRESS_MASKS[1:]):
        hold = m&x
        x &= ~m
        x >>= np.uint64(1<<i)
        x |= hold
    return x


def alt_expand(x, in_place=False):
    """
    Start in 0x00000000ffffffff state.

    https://help.dyalog.com/18.0/Content/Language/Primitive%20Functions/Expand.htm

    Examples
    --------
    >>> f'{alt_expand(0b100101):012b}'
    '010000010001'

    See Also
    --------
    alt_compress
    """

    o = len(COMPRESS_MASKS)
    if isinstance(x, int):
        x = np.uint64(x)
    elif not (in_place and np.issubdtype(x.dtype, np.uint64)):
        x = x.astype(np.uint64)
    x &= COMPRESS_MASKS[-1]
    for i, m in enumerate(COMPRESS_MASKS[-2::-1]):
        hold = m&x
        x &= ~m
        x <<= np.uint64(1<<(o-i-2))
        x |= hold
    return x



# Copyright (C) 2022-2023 Columbia Experimental Gravity Group (GECo)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Load skymaps from fits files.
"""

from __future__ import annotations

import re
import gzip
import logging
from pathlib import Path
from collections import OrderedDict
from typing import (
    Union,
    List,
    Tuple,
    Callable,
    Iterator,
    Any,
    Optional,
    Iterable,
    cast,
    TYPE_CHECKING
)
from ..healpy import healpy as hp
from ..utils import (
    set_partial_skymap_metadata,
    EmptyStream,
    nest2uniq,
    uniq2nside,
    nside2pixarea,
    uniq_minimize,
    uniq_intersection,
    is_gz,
)

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray

    UniqArray = NDArray['np.integer[Any]']

    from astropy.io.fits import (
        Header,
        BinTableHDU,
        ColDefs
    )
    from astropy.io.fits.hdu.base import ExtensionHDU
    from astropy.units.quantity import Quantity

    from _typeshed import SupportsRead

LOGGER = logging.getLogger(__name__)
FITS_BLOCKSIZE = 2880
FITS_CARD_WIDTH = 80
FITS_CARD_NAME_WIDTH = 8
FITS_HEADER_LENGTH = FITS_BLOCKSIZE // FITS_CARD_WIDTH
BINTABLE_TO_NUMPY_TYPES = {
    'L': '?',    # boolean
    # 'X': bit, not supported for now.
    'B': 'B',    # unsigned byte
    'I': '>i2',  # 16-bit big-endian int
    'J': '>i4',  # 32-bit big-endian int
    'K': '>i8',  # 64-bit big-endian int
    'A': 'b',    # char/signed byte
    'E': '>f4',  # 32-bit big-endian floating point
    'D': '>f8',  # 64-bit big-endian floating point
    'C': '>c8',  # 64-bit big-endian complex floating point
    'M': '>c16', # 128-bit big-endian complex floating point
    # 'P': 32-bit array descriptor, not supported for now
    # 'Q': 64-bit array descriptor, not supported for now
}
HEADER_NON_META = re.compile(
    "|".join(
        (
            'XTENSION',
            'BITPIX',
            'NAXIS[0-9]*',
            'PCOUNT',
            'GCOUNT',
            'TFIELDS',
            'TTYPE[0-9]*',
            'TFORM[0-9]*',
            'TUNIT[0-9]*',
        )
    )
)
TFORM = re.compile(r'([0-9]*)([LXBIJKAEDCMPQ])(.*)')
BUFFER_ROWS = 4**8


def calculate_max_rows_read(hdu, dtype, buf_rows):
    "Meant to handle BAYESTAR MOC and NEST + CWB skymaps."
    for name, coltype, *repeat in dtype:
        if name in ('PROBDENSITY', 'PROB', 'PROBABILITY'):
            if repeat:
                return buf_rows // repeat[0]
            return buf_rows
    raise ValueError(f"Could not find a PROBABILITY column in {hdu.columns}, "
                     "do not know how to proceed.")


def extract_probdensity(hdu, chunk, offset):
    "Meant to handle BAYESTAR MOC and NEST + CWB skymaps."
    import numpy as np
    from astropy.units import Quantity

    # define up here for typing purposes
    nside = -1

    ordering = hdu.header['ORDERING']
    if ordering == 'NUNIQ':
        u = chunk['UNIQ']
    else:
        for probname in ('PROBDENSITY', 'PROBABILITY', 'PROB'):
            if probname in hdu.columns.names:
                fmt = cast(str, hdu.columns[probname].format)
                m = TFORM.match(fmt)
                if m is None:
                    raise ValueError(f"Could not parse column format {fmt}")
                repeat = int(m[1]) if m[1] else 1
                break
        else:
            raise ValueError("Could not find a probability column in "
                             f"HDU columns: {hdu.columns}")
        nside = hdu.header['NSIDE']
        inds = np.arange(repeat*offset, repeat*(offset+len(chunk)))
        if ordering == 'RING':
            inds = hp.ring2nest(nside, inds)
        elif ordering != 'NESTED':
            raise ValueError(f"Unrecognized ordering: {ordering}")
        u = nest2uniq(inds, nside, in_place=True)
    if 'PROBDENSITY' in hdu.columns.names:
        return u, [Quantity(chunk['PROBDENSITY'], copy=False,
                            unit=hdu.columns['PROBDENSITY'].unit)]
    LOGGER.debug(f"PROBDENSITY not found in {hdu.columns}, trying to "
                 f"calculate it from PROB column and NSIDE")
    if 'PROBABILITY' in hdu.columns.names:
        LOGGER.debug("Column named PROBABILITY found. Using for probability "
                     "(Fermi GBM convention?)")
        probname = 'PROBABILITY'
    elif 'PROB' in hdu.columns.names:
        probname = 'PROB'
    else:
        raise ValueError(f"PROB not found in HDU columns: {hdu.columns}")
    probunit = hdu.columns[probname].unit
    if not probunit == 'pix-1':
        raise ValueError(f"Unexpected unit for {probname} column: {probunit}")
    prob = chunk[probname].ravel()  # handle CWB skymaps with repeated cols
    if ordering == 'NUNIQ':
        nside = uniq2nside(u)
    return u, [Quantity(prob/nside2pixarea(nside, degrees=False),
                        copy=False, unit='sr-1')]


def read_bintable_chunks(
        stream: SupportsRead[bytes],
        tables: int = -1,
        buf_rows: int = BUFFER_ROWS,
        extractor: Callable[
            ['BinTableHDU', 'NDArray[Any]', int],
            Tuple[
                'UniqArray',
                List['Quantity'],
            ]
        ] = extract_probdensity,
        row_calculator: Callable[
            [
                'BinTableHDU',
                List[Union[Tuple[str, str], Tuple[str, str, Tuple[int]]]],
                int,
            ],
            int
        ] = calculate_max_rows_read,
) -> Iterator[
        Tuple[
            'BinTableHDU',
            Iterator[
                Tuple[
                    'NDArray[Any]',
                    List['Quantity'],
                ]
            ],
        ]
]:
    import numpy as np
    from astropy.io import fits

    # assume first HDU has been skipped already
    while tables != 0:
        try:
            hdu = next_hdu(stream)
        except EmptyStream:
            break

        # make sure the HDUs are bintables
        assert isinstance(hdu, fits.BinTableHDU)

        dtype = bintable_dtype(hdu)
        max_rows = row_calculator(hdu, dtype, buf_rows)
        total_rows = cast(int, hdu.header['NAXIS2'])
        width = cast(int, hdu.header['NAXIS1'])
        assert total_rows * width == hdu.size

        # create a generator for each HDU
        def chunk_iter():
            offset = 0
            while offset < total_rows:
                rows = min(max_rows, total_rows - offset)
                chunk = np.frombuffer(stream.read(width*rows), dtype=dtype)
                yield extractor(cast('BinTableHDU', hdu), chunk, offset)
                offset += rows
            assert offset == total_rows
            position_in_block = hdu.size % FITS_BLOCKSIZE
            if position_in_block != 0:
                stream.read(FITS_BLOCKSIZE - position_in_block)

        yield hdu, chunk_iter()
        tables -= 1


def load_ligo(
        infile: Union[SupportsRead[bytes], str, Path],
        mask: Optional['UniqArray'] = None,
        maps: Optional[Union[int, Iterable[int]]] = None,
        extractor: Callable[
            ['BinTableHDU', 'NDArray[Any]', int],
            Tuple[
                'UniqArray',
                List['Quantity'],
            ]
        ] = extract_probdensity,
        row_calculator: Callable[
            [
                'BinTableHDU',
                List[Union[Tuple[str, str], Tuple[str, str, Tuple[int]]]],
                int,
            ],
            int
        ] = calculate_max_rows_read,
        chunk_processor: Optional[Callable[
            ['UniqArray', 'NDArray[Any]'],
            Tuple['UniqArray', 'NDArray[Any]']
        ]] = uniq_minimize,
        post_processor: Optional[Callable[
            ['UniqArray', 'NDArray[Any]'],
            Tuple['UniqArray', 'NDArray[Any]']
        ]] = uniq_minimize,
        buf_rows: int = BUFFER_ROWS,
) -> Iterator[Tuple['UniqArray', 'Quantity', OrderedDict]]:
    import numpy as np

    if isinstance(infile, (str, Path)):
        with (gzip.open if is_gz(infile) else open)(infile, 'rb') as stream:
            for tab in load_ligo(stream, mask, maps, extractor, row_calculator,
                                 chunk_processor, post_processor, buf_rows):
                yield tab
        return
    if mask is not None:
        mask, = uniq_minimize(mask)
    if maps is None:
        tables = -1
    elif isinstance(maps, int):
        tables = maps
        maps = None
    else:
        maps = np.array(list(maps))
        umaps = np.unique(maps)
        if len(umaps) != len(maps) or (umaps != maps).any():
            raise ValueError("Must provide a unique sorted list of `maps`.")
        del umaps
        tables = max(maps)
    # read the first HDU, which will be empty for a BINTABLE extension
    # fits file
    next_hdu(infile)
    for i, [hdu, table] in enumerate(
            read_bintable_chunks(
                infile,
                tables=tables,
                buf_rows=buf_rows,
                extractor=extractor,
                row_calculator=row_calculator
            )
    ):
        us = []
        ss = []
        for u, [s] in table:
            if maps is None or i in maps:
                if chunk_processor is not None:
                    u, s = chunk_processor(u, s)
                if mask is not None:
                    ui = np.unique(uniq_intersection(u, mask)[0])
                    us.append(u[ui])
                    ss.append(s[ui])
                else:
                    us.append(u)
                    ss.append(s)
        u = np.concatenate(us)
        s = np.concatenate(ss)
        del us, ss
        meta = OrderedDict()
        for k, v, *_ in hdu.header.cards:
            if HEADER_NON_META.match(k) is not None:
                continue
            if k == 'HISTORY':
                if k in meta:
                    meta[k].append(v)
                else:
                    meta[k] = [v]
            elif k not in meta:
                meta[k] = v
        set_partial_skymap_metadata(meta, mask, load_ligo.__name__)
        if post_processor is not None:
            u, s = post_processor(u, s)
        yield u, s, meta


def bintable_dtype(
        hdu: 'BinTableHDU',
) -> List[Union[Tuple[str, str], Tuple[str, str, Tuple[int]]]]:
    """
    Get list that can be passed to ``numpy.dtype`` to define a structured
    datatype corresponding to the data in the table ``hdu``.
    """
    types = []
    for col in cast('ColDefs', hdu.columns):
        m = TFORM.match(cast(str, col.format))
        assert m is not None
        g = m.groups()
        dt = BINTABLE_TO_NUMPY_TYPES[g[1]]
        tup = (col.name or '', dt)
        if g[0]:
            tup += (int(g[0]),)
        types.append(tup)
    return types


def _next_header_blocks(stream: SupportsRead[bytes]):
    block = stream.read(FITS_BLOCKSIZE)
    if not block:
        return block
    for i in range(FITS_HEADER_LENGTH):
        start = i*FITS_CARD_WIDTH
        if block[start:start+FITS_CARD_NAME_WIDTH] == b'END     ':
            return block
    return block + _next_header_blocks(stream)


def next_header(stream: SupportsRead[bytes]) -> 'Header':
    "Read the next ``astropy.fits.Header`` from a fits file stream (bytes)."
    from astropy.io import fits

    header = _next_header_blocks(stream)
    if not header:
        raise EmptyStream("Stream ended.")
    return fits.Header.fromstring(header)


def next_hdu(stream: SupportsRead[bytes]) -> 'ExtensionHDU':
    "Read the next FITS HDU from a fits file stream (bytes)."
    from astropy.io import fits

    header = _next_header_blocks(stream)
    if not header:
        raise EmptyStream("Stream ended.")
    return cast('ExtensionHDU', fits.HDUList.fromstring(header)[0])



# Copyright (C) 2021-2023 Columbia Experimental Gravity Group (GECo)

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Plotting commands implemented using Astropy (rather than HEALPy) for greater
flexibility, robustness, and cross-platform compatibility. Meant to replace
the plotters in ``plotters``. These provide similar functionality to the
plotters provided by the ``ligo.skymap`` package as well as ``healpy``; those
packages are not included as dependencies, but the high-performance rendering
tools included in ``hpmoc`` are totally compatible with those plotting tools
through the ``PartialUniqSkymap.render`` interface (in fact, the previous
version of these plotting scripts used ``healpy`` for its included projection
axes and coordinate transforms).

Available projections are drawn from the FITS standard. You can specify your
own projections using the appropriate FITS headers plugged into an
``astropy.wcs.WCS`` world coordinate system instance as the ``projection``
argument to a subplot or ``astropy.visualization.wcsaxes.WCSAxes`` instance.
The projection code follows the "4-3" form defined in the
`FITS Definition`_ document, i.e. by specifying the coordinate type
(e.g. ``RA--`` for right ascension) right-padded by enough dashes to fill
4 characters, followed by an additional ``-`` and the 3-character projection
code (one of those listed below) to specify the projection. This information is
then stored in the ``CTYPEi`` headers. Pixel-scaling through the ``CDELTi``
headers has a projection-dependent normalization and further depends on pixel
resolution of the final image. Details on WCS header fields are available in
the `FITS Definition`_, and the normalization factors can be extracted from
each projection's definition in
`Representations of celestial coordinates in FITS`_.

.. _`FITS Definition`: https://fits.gsfc.nasa.gov/fits_standard.html
.. _`Representations of celestial coordinates in FITS`: https://ui.adsabs.harvard.edu/abs/2002A&A...395.1077C

The complete list of available projections can be found in the
`FITS Definition`_, with concrete transformation definitions given in
the `Representations of Celestial Coordinates in FITS`_. The table of
available projections is reproduced below, with sections linking to the
corresponding section in `Representations of Celestial Coordinates in FITS`_.
Note that not all valid WCS projections can be displayed by astropy at time of
writing; in particular, the HEALPIX ``HPX`` projection does not work out of the
box, which is one of the (many) motivations for this plotting library.

.. list-table:: Available Projections
   :widths: 25 25 25 50 75
   :header-rows: 1

   * - Code
     - φ_0
     - θ_0
     - Properties1
     - Projection name
   * - AZP
     - 0◦
     - 90◦
     - Sect. 5.1.1
     - Zenithal perspective
   * - SZP
     - 0◦
     - 90◦
     - Sect. 5.1.2
     - Slant zenithal perspective
   * - TAN
     - 0◦
     - 90◦
     - Sect. 5.1.3
     - Gnomonic
   * - STG
     - 0◦
     - 90◦
     - Sect. 5.1.4
     - Stereographic
   * - SIN
     - 0◦
     - 90◦
     - Sect. 5.1.5
     - Slant orthographic
   * - ARC
     - 0◦
     - 90◦
     - Sect. 5.1.6
     - Zenithal equidistant
   * - ZPN
     - 0◦
     - 90◦
     - Sect. 5.1.7
     - Zenithal polynomial
   * - ZEA
     - 0◦
     - 90◦
     - Sect. 5.1.8
     - Zenithal equal-area
   * - AIR
     - 0◦
     - 90◦
     - Sect. 5.1.9
     - Airy
   * - CYP
     - 0◦
     - 0◦
     - Sect. 5.2.1
     - Cylindrical perspective
   * - CEA
     - 0◦
     - 0◦
     - Sect. 5.2.2
     - Cylindrical equal area
   * - CAR
     - 0◦
     - 0◦
     - Sect. 5.2.3
     - Plate carrée
   * - MER
     - 0◦
     - 0◦
     - Sect. 5.2.4
     - Mercator
   * - SFL
     - 0◦
     - 0◦
     - Sect. 5.3.1
     - Samson-Flamsteed
   * - PAR
     - 0◦
     - 0◦
     - Sect. 5.3.2
     - Parabolic
   * - MOL
     - 0◦
     - 0◦
     - Sect. 5.3.3
     - Mollweide
   * - AIT
     - 0◦
     - 0◦
     - Sect. 5.3.4
     - Hammer-Aitoff
   * - COP
     - 0◦
     - θa
     - Sect. 5.4.1
     - Conic perspective
   * - COE
     - 0◦
     - θa
     - Sect. 5.4.2
     - Conic equal-area
   * - COD
     - 0◦
     - θa
     - Sect. 5.4.3
     - Conic equidistant
   * - COO
     - 0◦
     - θa
     - Sect. 5.4.4
     - Conic orthomorphic
   * - BON
     - 0◦
     - 0◦
     - Sect. 5.5.1
     - Bonne’s equal area
   * - PCO
     - 0◦
     - 0◦
     - Sect. 5.5.2
     - Polyconic
   * - TSC
     - 0◦
     - 0◦
     - Sect. 5.6.1
     - Tangential spherical cube
   * - CSC
     - 0◦
     - 0◦
     - Sect. 5.6.2
     - COBE quadrilateralized spherical cube
   * - QSC
     - 0◦
     - 0◦
     - Sect. 5.6.3
     - Quadrilateralized spherical cube
   * - HPX
     - 0◦
     - 0◦
     - Sect. 6 2
     - HEALPix grid

See Also
--------
hpmoc.partial.PartialUniqSkymap
astropy.wcs.WCS
astropy.visualization.wcsaxes
matplotlib.axes.Axes
matplotlib.figure.Figure
ligo.skymap
healpy
"""

from __future__ import annotations

import math
from copy import deepcopy
from functools import partial
from warnings import warn
from typing import (
    Optional, Union, Tuple, Iterable, Callable,
    Sequence, Mapping, Sized, Any, TYPE_CHECKING
)
from textwrap import indent, wrap
from .healpy import healpy as hp
from .utils import (
    outline_effect,
    N_X_OFFSET,
    N_Y_OFFSET,
    wcs2ang,
    wcs2mask_and_uniq,
    monochrome_opacity_colormap,
)
from .points import PointsTuple

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    import astropy.io.fits
    from astropy.wcs.wcs import WCS
    import astropy.visualization.wcsaxes.frame

    import matplotlib.colors
    import matplotlib.gridspec
    import matplotlib.figure

    from .partial import PartialUniqSkymap

DEFAULT_CBAR_KWARGS = {
    'orientation': 'horizontal',
    'aspect': 40,
}
EDGE_EXCLUSION = 80
EXCLUSIONS_INITIALIZED = [False]
AIT_MOL_CDELT_BASE = math.sqrt(8) / math.pi
CEA_CDELT2_BASE = 2 / math.pi
TAN_CDELT_BASE = 0.025
SIN_CDELT_BASE = 2 / math.pi
ARC_CDELT_BASE = 2
ZEA_CDELT_BASE = 4 / math.pi
DEFAULT_WIDTH = 360
DEFAULT_HEIGHT = 180
DEFAULT_ROT = (180, 0, 0)
DEFAULT_FACING = True
DEFAULT_DPI = 200
DEFAULT_GRID_ROW_HEIGHT = 4
DEFAULT_HSPACE = 0.2
DEFAULT_WSPACE = 0.2
DEFAULT_NCOLS = 2
BASE_FONT_SIZE = 10
DEFAULT_C_KWARGS = {}
DEFAULT_CLABEL_KWARGS = {
    'inline': False,
    'fontsize': BASE_FONT_SIZE,
}
_DELTA_HIGHRES = tuple(v*.1**i for i in range(20) for v in (5, 2, 1))
DELTA_PARALLELS = (15.,)+_DELTA_HIGHRES     # space btwn graticule parallels
DELTA_MERIDIANS = (30., 10.)+_DELTA_HIGHRES #   and meridians [deg]
MIN_GRAT = 3
_WCS_HEADERS = dict()
_WCS_FRAMES = dict()
_SHARED = {
    "NAXIS": 2,
    "NAXIS2": 180,
    "CRVAL1": 180.0,
    "CTYPE1": 'RA---',
    "CUNIT1": 'deg',
    "CRVAL2": 0.0,
    "CTYPE2": 'DEC--',
    "CUNIT2": 'deg',
    "COORDSYS": 'icrs'
}
_ALL_SKY = _SHARED.copy()
_ALL_SKY["NAXIS1"] = 360.0
_ZENITHAL = _SHARED.copy()
_ZENITHAL["NAXIS1"] = 180.0
_ZENITHAL["LONPOLE"] = 180.0
_docs = {
    "allsky": "\n",
    "zenithal": "\n",
}
_PROJ_SETTINGS = {
    "allsky": (
        _ALL_SKY,
        {
            'MOL': (('Mollweide', 'mollview', 'Homolographic', 'Homalographic',
                     'Babinet', 'Elliptical'),
                    {'CDELT1': AIT_MOL_CDELT_BASE,
                     'CDELT2': AIT_MOL_CDELT_BASE}, 'e'),
            'AIT': (('Hammer-Aitoff', 'Aitoff', 'Hammer', 'Aitov',
                     'Hammer equal-area'),
                    {'CDELT1': AIT_MOL_CDELT_BASE,
                     'CDELT2': AIT_MOL_CDELT_BASE}, 'e'),
            'CAR': (('Carée', 'Plate carée', 'Caree', 'Plate caree',
                     'Cartesian', 'Tyre', 'cartview',
                     'Equidistant cylindrical'),
                    {'CDELT1': 1, 'CDELT2': 1}, 'r'),
            'CEA': (('Cylindrical equal-area', 'Lambert equal area'),
                    {'CDELT1': 1, 'CDELT2': CEA_CDELT2_BASE}, 'r'),
            'SFL': (('Sanson-Flamsteed',),
                    {'CDELT1': 1, 'CDELT2': 1}, 'p'),
            'PAR': (('Parabolic', 'Craster'),
                    {'CDELT1': 1, 'CDELT2': 1}, 'p'),
        },
    ),
    "zenithal": (
        _ZENITHAL,
        {
            # NB: TAN is just AZP (zenithal perspective) with mu set to zero.
            'TAN': (('gnomonic', 'gnomview', 'Central', 'zoom'),
                    {'CDELT1': TAN_CDELT_BASE, 'CDELT2': TAN_CDELT_BASE}, 'r'),
            'SIN': (('Slant Orthographic', 'Orthographic', 'Globe',
                     'orthview'),
                    {'CDELT1': SIN_CDELT_BASE, 'CDELT2': SIN_CDELT_BASE}, 'e'),
            'ARC': (('Zenithal Equidistant', 'azimuthal equidistant',
                     'azeqview', 'Postel', 'Equidistant', 'Globular'),
                    {'CDELT1': ARC_CDELT_BASE, 'CDELT2': ARC_CDELT_BASE}, 'e'),
            'ZEA': (('Zenithal Equal-area', 'Azimuthal equivalent',
                     'polar azimuthal', 'Lambert azimuthal equivalent',
                     'Lambert azimuthal equal-area',
                     'Lambert polar azimuthal', 'Lambert'),
                    {'CDELT1': ZEA_CDELT_BASE, 'CDELT2': ZEA_CDELT_BASE}, 'e'),
        },
    ),
}
# Add all-sky projections; aliases drawn from common usage, healpy.visufunc,
# WCS standard, and aliases as given in astro-ph/0207413
for docname, (defaults, config) in _PROJ_SETTINGS.items():
    for proj, (aliases, projdefaults, frame) in config.items():
        _docs[docname] += '        - ' + indent('\n'.join(
            wrap(f"{proj}: *{', '.join(aliases)}*", 69)), ' '*10)[10:] + '\n'
        _WCS_FRAMES[proj] = frame
        _WCS_HEADERS[proj] = defaults.copy()
        _WCS_HEADERS[proj]['CTYPE1'] += proj
        _WCS_HEADERS[proj]['CTYPE2'] += proj
        _WCS_HEADERS[proj].update(projdefaults)
        for alias in aliases:
            _WCS_HEADERS[alias.upper()] = _WCS_HEADERS[proj]
            _WCS_FRAMES[alias.upper()] = frame
    _docs[docname] = _docs[docname].rstrip('\n')


def _one_pt(scatter, rot):
    return len(scatter) == 1 and len(scatter[0].points) == 1 and rot is None


def _set_delts(projection, hdelta, vdelta, scatter, sigmas):
    return (projection in ['ARC', *_PROJ_SETTINGS['zenithal'][1]['ARC'][0]]
            and hdelta is None and vdelta is None and sigmas
            and len(scatter[0].points[0]) > 2)


def get_frame_class(
        projection: Union[
            str,
            'WCS',
            'astropy.io.fits.Header',
        ] = 'MOL',
        frame_class: Optional[
            Union[
                str,
                'astropy.visualization.wcsaxes.frame.BaseFrame',
            ]
        ] = None,
        vdelta: Optional[float] = None,
        hdelta: Optional[float] = None,
        rot: Optional[
            Union[
                Tuple[float, float, float],
                Tuple[float, float],
            ]
        ] = None,
        scatter: Sequence[PointsTuple] = tuple(),
        sigmas: Sequence[float] = (1,),
) -> 'astropy.visualization.wcsaxes.frame.BaseFrame':
    """
    Get the frame class associated with a given projection. If already given a
    ``frame_class``, returns it immediately, making this function idempotent.
    Will otherwise try to determine the best frame choice from the given
    arguments.

    Parameters
    ----------
    projection : str, WCS, or Header, optional
        The projection to use. See ``get_wcs`` for details. Only used if a
        ``str``.
    frame_class : str or BaseFrame
        The frame class; optionally pass it to ensure that an argument is an
        instance of ``BaseFrame``.
    vdelta : float, optional
        Manual vdelta override. Give up if set and return ``RectangularFrame``.
    hdelta : float, optional
        Manual hdelta override. Give up if set and return ``RectangularFrame``.
    rot, scatter, sigmas
        See the note in ``get_wcs`` on these arguments. If both rotation and
        window are set for ARC (zenithal equidistant), switch to
        ``RectangularFrame``.

    Returns
    -------
    frame_class: astropy.visualization.wcsaxes.frame.BaseFrame
        The frame class best-suited to this type of projection.

    Raises
    ------
    IndexError
        If the specified projection could not be found.
    """
    from astropy.visualization.wcsaxes import frame

    if isinstance(frame_class, type) and \
            issubclass(frame_class, frame.BaseFrame):
        return frame_class
    if isinstance(frame_class, str):
        return {
            'ELLIPTICAL': frame.EllipticalFrame,
            'RECTANGULAR': frame.RectangularFrame,
        }[frame_class.upper()]
    if {vdelta, hdelta} != {None} or not isinstance(projection, str):
        return frame.RectangularFrame
    if _one_pt(scatter, rot) and _set_delts(projection, hdelta, vdelta,
                                            scatter, sigmas):
        return frame.RectangularFrame
    f = _WCS_FRAMES[projection.upper()]
    if f == 'e':
        return frame.EllipticalFrame
    if f == 'r':
        return frame.RectangularFrame
    warn(f"Frame class for {projection} not yet available. Using default for "
         "now; specify your own for a better-looking plot.", UserWarning)
    return frame.RectangularFrame


GET_FRAME_CLASS_KWARG_KEYS = ('frame_class', 'vdelta', 'hdelta', 'rot',
                              'scatter', 'sigmas')
GET_WCS_KWARG_KEYS = ('width', 'height', 'hdelta', 'vdelta', 'rot',
                      'facing_sky', 'scatter', 'sigmas')


# TODO allow ICRS override
def get_wcs(
        projection: str,
        width: Optional[int] = None,
        height: int = DEFAULT_HEIGHT,
        hdelta: Optional[float] = None,
        vdelta: Optional[float] = None,
        rot: Optional[
            Union[
                Tuple[float, float, float],
                Tuple[float, float],
            ]
        ] = None,
        facing_sky: bool = DEFAULT_FACING,
        scatter: Sequence[PointsTuple] = tuple(),
        sigmas: Sequence[float] = (),
) -> 'WCS':
    """
    Get a ``WCS`` instance by name to match the given parameters.

    Parameters
    ----------
    projection: str
        The following projections are available by default. See ``hpmoc.plot``
        documentation for instructions on constructing your own WCS headers for
        other plot styles, which can be passed in instead using a ``WCS``
        instance. You can also use this approach to plot the skymap over an
        existing ``WCS`` taken from another fits file, making it easy to plot
        skymaps over other astrophysical data. This function will return
        ``WCS`` instances for the following projection types:

        **Cylindrical/Pseudo-cylindrical** (For all-sky plots)
        {allsky}

        **Zenithal**
        {zenithal}
    width: int, optional
        Width of the image in pixels. If not provided, default to the height
        for zenithal projections and twice the height for azimuthal
        projections.
    height: int
        Height of the image in pixels.
    hdelta: float, optional
        The CDELT1 value, if you wish to override the default. Note that the
        actual angular width of a pixel at the reference point depends on this
        value *as well as* the projection used. *Ignored if* ``ax`` *is given.*
    vdelta: float, optional
        The CDELT2 value, if you wish to override the default. Note that the
        actual angular height of a pixel at the reference point depends on this
        value *as well as* the projection used. *Ignored if* ``ax`` *is given.*
    rot: float, float, float) or (float, float), optional
        Euler angles for rotations about the Z, X, Z axes. These are
        immediately translated to ``CRVAL1, CRVAL2, LONPOLE`` in the returned
        ``WCS``; that is, the first two angles specify the angle of the center
        of the image, while the last specifies an additional rotation of the
        reference pole about the Z-axis. See ``hpmoc.plot`` documentation for
        further references. Defaults to centering on RA = 180, dec = 0 (the
        convention used in most plots produced by LIGO). If only two angles are
        provided, ``LONPOLE`` will be determined automatically such that the
        center of the skymap is given by the `(RA, dec)` given and the
        orientation of the plot is otherwise unchanged.
    facing_sky: bool
        Whether the projection is outward- or inward-facing. Equivalent to
        reversing the direction of the longitude.
    scatter : Sequence[PointsTuple]
        A list of collections of point sources that will be plotted. If only
        one collection containing one point source is provided and ``rot=None``
        (the default), then ``rot`` will be set to the location of that single
        point source.
    sigmas : Sequence[float]
        See ``plot`` for meaning. If ``sigmas`` is non-empty; ``rot`` is set by
        a single point source in ``scatter`` as described above; ``hdelta``
        and ``vdelta`` are both ``None``; and the ``projection`` is an alias
        for ARC (zenithal equidistant), then ``hdelta`` and ``vdelta`` will be
        set so that the returned frame is 1.5x the size of the largest error
        region plotted in the smallest axis. This provides an easy way to
        visually emphasize singular points of interest.

    Returns
    -------
    wcs: astropy.wcs.WCS
        A ``WCS`` instance that can be used for rendering and plotting.

    Raises
    ------
    IndexError
        If the specified projection could not be found or if an argument is
        specified incorrectly.
    """
    from astropy.io.fits import Header
    from astropy.wcs import WCS

    dec_dir = -1 if facing_sky else 1
    header = Header(_WCS_HEADERS[projection.upper()].copy())
    if width is not None:
        header['CDELT1'] *= header['NAXIS1'] / width # type: ignore
        header['NAXIS1'] = width # type: ignore
    header['CRPIX1'] = header['NAXIS1'] / 2 + 0.5 # type: ignore
    header['CDELT2'] *= header['NAXIS2'] / height # type: ignore
    header['NAXIS2'] = height # type: ignore
    header['CRPIX2'] = header['NAXIS2'] / 2 + 0.5 # type: ignore
    # rot set to point location if only one point present
    if _one_pt(scatter, rot):
        pt = scatter[0].points[0]
        rot = tuple(pt[:2])
        if _set_delts(projection, hdelta, vdelta, scatter, sigmas):
            sig = 1.5 * max(sigmas) * pt[2] # type: ignore
            header['CDELT1'] *= sig / 180
            header['CDELT2'] *= sig / 180
    if rot:
        header['CRVAL1'] = rot[0]
        header['CRVAL2'] = rot[1]
        if len(rot) == 3:
            header['LONPOLE'] = rot[2]
        elif len(rot) == 2:
            if rot[1] < 0:
                for name, (aliases, *_) in _PROJ_SETTINGS['allsky'][1].items():
                    if projection.upper() in [n.upper()
                                              for n in [name, *aliases]]:
                        header['LONPOLE'] = 180
                        break
        else:
            raise ValueError(f"rot must have len 2 or 3; got {rot}")
    if hdelta is not None:
        header['CDELT1'] = hdelta
    if vdelta is not None:
        header['CDELT2'] = vdelta
    header['CDELT1'] *= dec_dir # type: ignore
    return WCS(header)


get_wcs.__doc__ = get_wcs.__doc__.format(allsky=_docs['allsky'], # type: ignore
                                         zenithal=_docs['zenithal'])


def get_colormap(cmap, bad=None):
    from matplotlib.cm import get_cmap
    from matplotlib.colors import Colormap

    if not isinstance(cmap, Colormap):
        cmap = deepcopy(get_cmap(cmap))
    if bad is not None:
        cmap.set_bad(color=bad)
    return cmap


def get_projection(projection, *args, **kwargs):
    """
    If ``projection`` is already an ``astropy.wcs.WCS`` instance, return it; if
    it is an ``astropy.io.fits.Header``, return ``WCS(projection)``. If it is a
    string, try to create a new ``WCS`` using ``get_wcs`` with ``*args`` and
    ``**kwargs`` passed to that function; failing that, try to interpret
    ``projection`` as a raw FITS header and instantiate a ``WCS`` therefrom.
    """
    from astropy.wcs import WCS
    from astropy.io.fits import Header

    if isinstance(projection, WCS):
        return projection
    if isinstance(projection, Header):
        return WCS(projection)
    try:
        return get_wcs(projection, *args, **kwargs)
    except IndexError:
        return WCS(Header.fromstring(projection))


# TODO allow ICRS override
def plot(
        skymap: Union[
            'PartialUniqSkymap',
            NDArray[Any],
            Tuple[
                NDArray[Any],
                Optional[
                    Union[
                        NDArray['np.integer[Any]'],
                        'WCS',
                        str,
                    ]
                ],
            ],
        ],
        *scatter: PointsTuple,
        scatter_marker_size: float = BASE_FONT_SIZE*4,
        scatter_label_size: float = BASE_FONT_SIZE,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        cmap: Optional[
            Union[str, 'matplotlib.colors.Colormap']
        ] = 'gist_heat_r',
        missing_color: Optional[Union[str, Tuple[int, int, int]]] = None,
        nan_color: Optional[Union[str, Tuple[int, int, int]]] = None,
        alpha: float = 1.,
        sigmas: Sequence[float] = [],
        outline_sigmas: Sequence[float] = [],
        scatter_labels: Union[bool, dict] = True,
        ax: Optional[Union[
            'astropy.visualization.wcsaxes.WCSAxes',
            'astropy.visualization.wcsaxes.WCSAxesSubplot'
        ]] = None,
        projection: Union[
            str,
            'WCS',
            'astropy.io.fits.Header',
        ] = 'Mollweide',
        frame_class: Optional[
            Union[
                'astropy.visualization.wcsaxes.frame.BaseFrame',
                str
            ]
        ] = None,
        width: Optional[int] = None,
        height: int = DEFAULT_HEIGHT,
        hdelta: Optional[float] = None,
        vdelta: Optional[float] = None,
        rot: Optional[
            Union[
                Tuple[float, float, float],
                Tuple[float, float],
            ]
        ] = None,
        facing_sky: bool = DEFAULT_FACING,
        fig: Optional[Union['matplotlib.figure.Figure', dict]] = None,
        subplot: Optional[
            Union[
                Tuple[int, int, int],
                'matplotlib.gridspec.SubplotSpec',
            ]
        ] = None,
        cr: Iterable[float] = tuple(),
        cr_format: Optional[Callable[[float, float], str]] = None,
        cr_filled: bool = False,
        cr_kwargs: Optional[dict] = None,
        cr_label_kwargs: Optional[dict] = None,
        pixels: Union[bool, dict] = False,
        pixels_format: Optional[
            Callable[['PartialUniqSkymap'], Iterable[str]]
        ] = None,
        cbar: Union[bool, dict] = False,
) -> Union[
    'astropy.visualization.wcsaxes.WCSAxes',
    'astropy.visualization.wcsaxes.WCSAxesSubplot'
]:
    """
    Parameters
    ----------
    skymap : 'PartialUniqSkymap', array, or (array, array)
        The skymap to plot. Can be a ``PartialUniqSkymap``, a single-resolution
        HEALPix skymap *in NEST ordering only*, or a tuple of (pixel values,
        NUNIQ indices) accepted as the first two arguments to
        ``PartialUniqSkymap``.
    scatter: PointsTuple
        Point-sources to plot as a scatter-map, with disks showing their error
        regions. Provide multiple (ideally with different colors) to plot many
        populations at once.
    scatter_marker_size: float, optional
        Point-source marker font size.
    scatter_label_size: float, optional
        Point-source label font size.
    vmin: float, optional
        The smallest value in the color map used to plot ``skymap``. Set
        ``None`` to have it calculated automatically.
    vmax: float, optional
        The largest value in the color map used to plot ``skymap``. Set
        ``None`` to have it calculated automatically.
    cmap: str or matplotlib.colors.Colormap
        The color map to use to plot ``skymap``. Note that the colors for the
        point sources in ``scatter`` are set using the ``rgba`` parameter in
        ``PointsTuple`` and will not be affected by this value. If ``None``,
        the skymap itself will not be plotted; this can be useful if overlaying
        multiple skymaps.
    missing_color : str or (int, int, int), optional
        The color to use for missing parts of the skymap. If not provided, they
        simply will not be shown.
    nan_color : str or (int, int, int), optional
        The color to use for parts of the map that are included but equal to
        ``np.nan``. If not provided, will be the same as ``missing_color``
        (meaning transparent if that argument is not provided), i.e. they will
        be plotted as if the pixels were missing. **Will not use the**
        ``alpha`` **argument (to allow for highlighting NaN values more
        clearly;** to use the same ``alpha`` value, you can create an RGBA
        tuple including the desired color and your shared alpha value, e.g.
        ``nan_color=[*matplotlib.colors.to_rgb('pink'), alpha]``.
    alpha: float, optional
        The opacity of the plotted skymap image.
    sigmas: Iterable[float], optional
        The size of the error region about each point source to plot in units
        of its error parameter sigma.
    outline_sigmas: Iterable[float], optional
        The size of the error region about each point source to plot as an
        unfilled circle in units of its error parameter sigma.
    scatter_labels: bool, optional
        Whether to show labels for the scattered points. If ``True``, display
        either their labels (if defined) or their indices within the
        ``PointsTuple.points`` list. If given as a ``dict``, will be
        interpreted as keyword arguments to pass to ``WCSAxes.text``, which can
        be used to control the appearance of the point source labels.
    ax: WCSAxes or WCSAxesSubplot, optional
        Axes to plot to. If provided, all other arguments pertaining to
        creating a ``WCS`` and ``WCSAxes`` instance are ignored, and these axes
        are used instead.
    projection: str, WCS, or Header, optional
        Either provide the name of the projection (see ``get_wcs`` docstring
        for valid names) to create a new ``WCS`` for this plot, or provide
        a ready-made ``WCS`` instance or FITS ``Header`` from which such an
        instance can be crafted. In the first case, you will need to specify
        other parameters needed to fully define the world coordinate system.
        In the latter cases, you might need to customize ``frame``.
        *Ignored if* ``ax`` *is given.*
    frame_class: BaseFrame or str, optional
        The frame type to use for this plot, e.g. a ``RectangularFrame`` (for
        a plate carée/Cartesian plot) or an ``EllipticalFrame`` for a
        Mollweide plot. Selected automatically when ``projection`` is specified
        by name, otherwise defaults to ``RectangularFrame``. You can also
        specify `frame_class='rectangular'` or `frame_class='elliptical'` to
        choose one of these two frames.
    width: int, optional
        Width of the image in pixels. If not provided, default to the height
        for zenithal projections and twice the height for azimuthal
        projections. *Ignored if* ``ax`` *is given or if* ``projection`` *is a*
        ``WCS`` *instance.*
    height: int, optional
        The height of the plot in pixels. *Ignored if* ``ax`` *is given or if*
        ``projection`` *is a* ``WCS`` *instance.*
    hdelta: float, optional
        The CDELT1 value, if you wish to override the default. Note that the
        actual angular width of a pixel at the reference point depends on this
        value *as well as* the projection used. *Ignored if* ``ax`` *is given
        or if* ``projection`` *is a* ``WCS`` *instance.*
    vdelta: float, optional
        The CDELT2 value, if you wish to override the default. Note that the
        actual angular height of a pixel at the reference point depends on this
        value *as well as* the projection used. *Ignored if* ``ax`` *is given
        or if* ``projection`` *is a* ``WCS`` *instance.*
    rot: (float, float, float) or (float, float), optional
        Euler angles for rotations about the Z, X, Z axes. These are
        immediately translated to ``CRVAL1, CRVAL2, LONPOLE`` in the returned
        ``WCS``; that is, the first two angles specify the angle of the center
        of the image, while the last specifies an additional rotation of the
        reference pole about the Z-axis. See ``hpmoc.plot`` documentation for
        further references. Defaults to centering on RA = 180, dec = 0 (the
        convention used in most plots produced by LIGO). If only two angles are
        provided, ``LONPOLE`` will be determined automatically such that the
        center of the skymap is given by the `(RA, dec)` given and the
        orientation of the plot is otherwise unchanged. *Ignored if* ``ax``
        *is given or if* ``projection`` *is a* ``WCS`` *instance.*
    facing_sky: bool, optional
        Whether the projection is outward- or inward-facing. Equivalent to
        reversing the direction of the longitude. *Ignored if* ``ax`` *is given
        or if* ``projection`` *is a* ``WCS`` *instance.*
    fig: matplotlib.figure.Figure or dict, optional
        The figure to plot to. If not provided, a new figure will be created.
        If a dictonary is provided, it will be passed as keyword arguments to
        create a new figure. *Ignored if* ``ax`` *is given.*
    subplot: (int, int, int) or SubplotSpec, optional
        If provided, initialize the plot as a subplot using the standard
        ``Figure.subplot`` matplotlib interface, returning a ``WCSAxesSubplot``
        instance rather than a ``WCSAxes`` instance.
        *Ignored if* ``ax`` *is given.*
    cr: Iterable[float], optional
        If provided, plot contour lines around the credible regions
        specified in this list. For example, ``cr=[0.9]`` will plot contours
        around the smallest region containing 90% of the skymap's integrated
        value.
    cr_format: Callable[[float, float], str], optional
        A function taking the CR level (e.g. ``0.9`` for 90% CR) and the actual
        value of the skymap on that contour and returning a string to be used
        to label each contour. If not provided, contours will not be labeled.
        *Ignored if* ``cr`` *is empty.*
    cr_filled: bool, optional
        Whether to fill the contours using ``contourf`` rather than
        ``contour``. *Ignored if* ``cr`` *is empty.*
    cr_kwargs: dict, optional
        Additional arguments to pass to either ``contour`` or ``contourf``.
        *Ignored if* ``cr`` *is empty.*
    cr_label_kwargs: dict, optional
        Arguments to pass to ``clabel`` governing the display format of
        contour labels.
    pixels: bool or dict, optional
        Whether to plot pixel borders. *You should probably keep this*
        ``False`` *if you are doing an all-sky plot, since border plotting is
        slow for a large number of pixels, and the boundaries will not be
        visible anyway unless each visible pixel's size is comparable to the
        overall size of the plot window.* If you just want to see information
        about e.g. the size of pixels across the whole sky, consider plotting
        ``skymap.orders(as_skymap=True)`` to see a color map of pixel sizes
        instead. If given as a ``dict``, will be interpreted as keyword
        arguments to pass to ``WCSAxes.plot``, which can be used to control the
        appearance of the pixel borders.
    pixels_format: Callable[[PartialUniqSkymap], Iterable[str]], optional
        A function that takes a skymap and returns a string for each pixel.
        Will be called on the selection of pixels overlapping with the visible
        area, so don't worry about optimizing it, since just plotting the
        borders will be slow enough for a large number of visible pixels.
        The returned string will be plotted over the center of each pixel,
        which again is only useful if the pixels are large with respect to the
        size of the plot window.
    cbar: bool or dict, optional
        If ``True``, add a colorbar for the plotted skymap. If a ``dict`` is
        provided, pass it as the keyword arguments to
        ``matplotlib.pyplot.colorbar``. *Ignored if* ``cmap=None``.

    Returns
    -------
    ax: WCSAxes or WCSAxesSubplot
        The axes that were just plotted to.

    See Also
    --------
    hpmoc.partial.PartialUniqSkymap
    matplotlib.figure.Figure
    matplotlib.axes.Axes
    matplotlib.pyplot.colorbar
    matplotlib.gridspec.GridSpec
    matplotlib.gridspec.SubplotSpec
    astropy.io.fits.Header
    astropy.wcs.WCS
    astropy.visualization.wcsaxes.WCSAxes
    """
    import numpy as np
    from matplotlib import pyplot as plt
    from matplotlib.figure import Figure
    from matplotlib.transforms import ScaledTranslation
    from matplotlib.gridspec import SubplotSpec
    from astropy.coordinates.sky_coordinate import SkyCoord
    from astropy.visualization.wcsaxes import WCSAxes, SphericalCircle
    from astropy.visualization.wcsaxes.frame import (
        RectangularFrame,
        EllipticalFrame,
    )
    from astropy.units import Quantity, deg
    from .partial import PartialUniqSkymap

    # initialize arguments
    cr = np.unique([*cr, 1])
    if not isinstance(skymap, PartialUniqSkymap):
        if isinstance(skymap, np.ndarray):
            skymap = (skymap, None)
        skymap = PartialUniqSkymap(*skymap)
    ## rot set to point location if only one point present
    #if len(scatter) == 1 and len(scatter[0].points) == 1 and rot is None:
    #    rot = tuple(scatter[0].points[0][:2])

    # initialize our axes if needed
    if ax is None:
        # initialize figure (just to get our axes)
        if not isinstance(fig, Figure):
            fig = plt.figure(**(fig or {}))
        # initialize frame class
        frame_class = get_frame_class(projection, frame_class, vdelta, hdelta,
                                      rot, scatter, sigmas)
        # initialize projection
        projection = get_projection(projection, width=width, height=height,
                                    hdelta=hdelta, vdelta=vdelta, rot=rot,
                                    facing_sky=facing_sky, scatter=scatter,
                                    sigmas=sigmas)
        # initialize axes
        if subplot is None:
            ax = WCSAxes(fig, rect=[0.1, 0.1, 0.9, 0.9], wcs=projection,
                         frame_class=frame_class)
            fig.add_axes(ax)
        else:
            _subplot = subplot
            if isinstance(_subplot, SubplotSpec):
                _subplot = (subplot,)
            ax = fig.add_subplot(*_subplot, projection=projection, # type: ignore
                                 frame_class=frame_class)
    else:
        # initialize projection
        projection = ax.wcs

    assert ax is not None
    # get the pixel coordinates
    _, ra, dec = wcs2ang(ax.wcs)

    # set default coordinate ticks and style
    outline = [outline_effect()]
    co_ra, co_dec = ax.coords
    ra_ticks = get_ticks(ra.value+3, ra_exclusions(ax)+3,
                         np.arange(30, 390, 30), 3, lambda x, d: x+d%360)
    co_ra.set_major_formatter("dd")
    if len(ra_ticks) >= MIN_GRAT:
        co_ra.set_ticks(ra_ticks * deg)
        #co_ra.set_ticks(spacing=30 * deg)
    else:
        # use degree formatting by default
        filled_arcmin = (np.histogram(ra, bins=360*120)[0] != 0).sum()
        if filled_arcmin < 10:
            co_ra.set_major_formatter("dd:mm:ss")
        elif filled_arcmin < 60 * 10:
            co_ra.set_major_formatter("dd:mm")
    dec_ticks = get_ticks(dec.value, dec_exclusions(ax),
                          np.arange(-75, 90, 15), 2, lambda x, _: x)
    if len(dec_ticks) >= MIN_GRAT:
        co_dec.set_major_formatter("dd")
        co_dec.set_ticks(dec_ticks * deg)
    # ticks can look bad on EllipticalFrame
    if issubclass(ax.frame_class, EllipticalFrame):
        co_ra.set_ticks_visible(False)
        co_dec.set_ticks_visible(False)
    co_ra.set_ticklabel(size=BASE_FONT_SIZE, path_effects=outline)
    co_dec.set_ticklabel(size=BASE_FONT_SIZE, path_effects=outline)
    # TODO override if not ICRS
    co_ra.set_axislabel("Right ascension (ICRS) [deg]",
                        size=BASE_FONT_SIZE, path_effects=outline)
    co_dec.set_axislabel("Declination (ICRS) [deg]",
                         size=BASE_FONT_SIZE, path_effects=outline)
    ax.grid(True)

    # plot skymap
    if (cmap is not None) or (cr.size > 1):
        mask, u = wcs2mask_and_uniq(projection)
        render = skymap.render(projection, pad=np.nan, mask_missing=True)
    if cmap is not None:
        cmap = get_colormap(cmap, bad=missing_color)
        img = ax.imshow(render, vmin=vmin, vmax=vmax, cmap=cmap, alpha=alpha)
        if nan_color is not None:
            ncmap = monochrome_opacity_colormap("NanCMap", nan_color)
            if missing_color is not None:
                ncmap.set_bad(missing_color)
            nans = np.isnan(render)
            ax.imshow(nans, vmin=0, vmax=1, cmap=ncmap, alpha=alpha)
        if cbar is not False:
            kw = DEFAULT_CBAR_KWARGS.copy()
            label = skymap.name or ''
            if skymap.unit is not None:
                label += (' [{}]' if label else '{}').format(skymap.unit)
            if label:
                kw['label'] = label
            if cbar is not True:
                kw.update(cbar)
            plt.colorbar(img, ax=ax, **kw)

    # plot scatterplots, layering sigma regions first
    # TODO do not scatter plot or label points that are off of the plot axes
    transform = ax.get_transform('world')
    label_transform = transform + ScaledTranslation(N_X_OFFSET/2, N_Y_OFFSET/2,
                                                    ax.figure.dpi_scale_trans)
    for pts in scatter:
        # Skip pts if it is empty, since you cannot have empty SkyCoords
        if not pts.points:
            continue

        # only include points in the visible region
        pts_x, pts_y = ax.wcs.world_to_pixel(SkyCoord(
            *[*zip(*((r, d) for r, d, *_ in pts.points))]*deg, frame='icrs'
        ))
        include = (pts_x < ax.wcs.pixel_shape[0]+.5) & (pts_x > .5)
        include &= (pts_y < ax.wcs.pixel_shape[1]+.5) & (pts_y > .5)
        col = pts.rgba.to_hex(False)
        ax.scatter(pts_x[include], pts_y[include], c=col, marker=pts.marker,
                   s=scatter_marker_size, label=pts.label, path_effects=outline,
                   zorder=10)

        # plot the shaded error circles, if needed
        cm = pts.cmap()
        for sigma in sigmas:
            ax.imshow(pts.render(projection, extent=sigma), vmin=0, vmax=1,
                      cmap=cm, zorder=5)

        # plot the error circle outlines, if needed
        for outline_sigma in outline_sigmas:
            for ra, dec, sigma, _ in pts.points:
                coord = SkyCoord(ra * deg, dec * deg)
                s = SphericalCircle(coord, sigma * outline_sigma * deg,
                    edgecolor=col, facecolor='none',
                    transform=ax.get_transform('icrs'), zorder=5)
                ax.add_patch(s)

        # plot the scatter labels, if needed
        if scatter_labels:
            for i, (r, d, *sl) in enumerate(pts.points):
                if not include[i]:
                    continue
                if len(sl) == 2:
                    pt_label = sl[1]
                else:
                    pt_label = str(i+1)
                kw = {
                    'fontsize': scatter_label_size,
                    'path_effects': outline,
                    'color': col,
                    'va': 'bottom',
                    'ha': 'left',
                    'zorder': 15
                }
                if not isinstance(scatter_labels, bool):
                    kw.update(scatter_labels)
                ax.text(r, d, pt_label, transform=label_transform, **kw)

    # plot contours
    if cr.size > 1:
        q = (1-cr)[::-1]
        _, levels, _ = skymap.quantiles((1-cr[::-1]))
        levels = levels[1:]
        cr_lut = dict(
            zip(
                levels.value if isinstance(levels, Quantity) else levels,
                cr[-2::-1]
            )
        )
        ptrans = ax.get_transform(projection)
        # TODO handle contourf needing two levels and reconcile it with
        # disposing of the 1.0 quantile
        contour = ax.contourf if cr_filled else ax.contour
        ckw = DEFAULT_C_KWARGS.copy()
        if cr_kwargs is not None:
            ckw.update(cr_kwargs)
        cntrs = contour(render, transform=ptrans, levels=levels, **ckw)
        if cr_format is not None:
            clabelkw = DEFAULT_CLABEL_KWARGS.copy()
            if cr_label_kwargs is not None:
                clabelkw.update(cr_label_kwargs)
            ax.clabel(cntrs, cntrs.levels, **clabelkw,
                      fmt=lambda v: cr_format(cr_lut[v], v))

    return ax


def gridplot(
        *skymaps: Union[
            'PartialUniqSkymap',
            NDArray[Any],
            Tuple[
                NDArray[Any],
                Optional[
                    Union[
                        NDArray['np.integer[Any]'],
                        'WCS',
                        str,
                    ]
                ],
            ],
        ],
        fig: Optional[
            Union[
                'matplotlib.figure.Figure',
                'matplotlib.gridspec.GridSpec',
                Mapping,
            ]
        ] = None,
        projections: Sequence[
            Union[
                str,
                'WCS',
                'astropy.io.fits.Header',
                Sequence[
                    Union[
                        str,
                        'WCS',
                        'astropy.io.fits.Header',
                        'astropy.visualization.wcsaxes.WCSAxes',
                    ]
                ],
            ]
        ]= ('MOL',),
        scatters: Optional[Sequence[Sequence[PointsTuple]]] = None,
        # fig args
        subplot_height: float = DEFAULT_GRID_ROW_HEIGHT,
        # fig.add_gridspec args
        ncols: int = DEFAULT_NCOLS,
        hspace: float = DEFAULT_HSPACE,
        wspace: float = DEFAULT_WSPACE,
        wshrink: Union[float, Sequence[float]] = 1.,
        # plotter args
        subplot_kwargs: Optional[Sequence[Optional[Sequence[dict]]]] = None,
        left=0.,
        right=1.,
        bottom=0.,
        top=1.,
        **kwargs
) -> Tuple[
        'matplotlib.gridspec.GridSpec',
        Sequence[Sequence['astropy.visualization.wcsaxes.WCSAxes']]
]:
    """
    Make a grid plot of multiple skymaps (optionally with scatterplots for
    each).

    Parameters
    ----------
    *skymaps : 'PartialUniqSkymap', array, or (array, array)
        The skymaps to plot. Can be a ``PartialUniqSkymap``, a
        single-resolution HEALPix skymap *in NEST ordering only*, or a tuple of
        (pixel values, NUNIQ indices) accepted as the first two arguments to
        ``PartialUniqSkymap``.
    fig: matplotlib.figure.Figure or dict, optional
        The figure to plot to. If not provided, a new figure will be created.
        If a dictonary is provided, it will be passed as keyword arguments to
        create a new figure. If a ``GridSpec`` is provided, then the figure
        to which it is attached will be used, and that ``GridSpec`` will be used
        to define the layout.
    projections : Sequence[Union[str, WCS, fits.Header, Sequence[Union[str, WCS, fits.Header, WCSAxes]]]], optional
        A list of projections (see the ``projection`` argument of ``plot``) to
        use for each skymap in ``skymaps``. If multiple projections are
        specified, they will be plotted alongside each other; if this makes the
        figure too wide, change the number of columns in the grid with
        ``ncols``. You can also pass a list of lists of axes of the type
        returned by this function (which you should do while also passing a
        ``GridSpec`` as ``fig``), allowing you to plot multiple layers of data
        to the same grid plot.
    scatters : Sequence[Sequence[PointsTuple]], optional
        Scatterplots to use, one list for each skymap containing the sets of
        points to plot for that skymap. For any of the skymaps which are
        ``PartialUniqSkymap`` instances, you can default to plotting that
        instance's ``point_sources`` by passing ``None`` instead of a list of
        point sources; this is also the default behavior if ``scatters=None``.
    subplot_height : float, optional
        The height of each subplot (in inches). Width is automatically
        determined. Overall figure height will be this times ``ncols``.
        *Ignored if* ``fig`` *is a pre-existing figure, or if* ``figsize`` *is
        specified as a keyword argument in* ``fig`` (though this is not
        recommended usage and should be avoided unless you know what you're
        doing).
    ncols : int, optional
        How many columns of *skymaps* to include in the grid. **NB: All
        subplots for a single skymap are counted as a single column** (in
        contrast to ``matplotlib.gridspec.GridSpec``, which counts each subplot
        in a single column). The number of rows is determined automatically
        from the number of skymaps provided combined with ``ncols``.
    hspace : float, optional
        How much vertical space (height) to reserve between subplots.
    wspace : float, optional
        How much horizontal space (width) to reserve between subplots.
    wshrink : float or list of floats, optional
        Scale the plot widths by this much. Useful if you are adding color
        bars to preserve spacing. If a list, each element corresponds to a
        projection.
    subplot_kwargs : Sequence[Optional[Sequence[dict]]], optional
        Lists of keyword argument dictionaries that will be used for each
        subplot. The first index specifies the plotter, and the second index
        specifies the skymap from ``skymaps``. This behavior allows you to
        easily specify lists of keyword arguments for specific projections
        (since often) only one of the ``projections`` requires skymap-specific
        parameters).  **NB: These subplot-specific keyword arguments take
        precedence over** ``**kwargs`` **for their respective subplots.
        Projection-related keyword arguments are ignored if** ``projections``
        **is a list of lists of axes; see the** ``plot`` **documentation
        for further details.**
    left, right, bottom, top : float
        Bounds for the GridSpec within the plot. Move ``top`` down, for
        example, if you want more space for a super-title with
        ``plt.suptitle``.
    **kwargs
        Keyword arguments applied to all projections; **again, see** ``plot``
        **for usages and caveats.**

    Returns
    -------
    gs : matplotlib.gridspec.GridSpec
        The ``GridSpec`` defining the shape of the subplots. Access the plotted
        figure as ``gs.figure``. Reuse this layout by passing ``gs`` to another
        invocation of ``gridplot``.
    axs : Sequence[Sequence['astropy.visualization.wcsaxes.WCSAxes']]
        The axes which were plotted. The first index (outer list) corresponds
        to the projection used, and the second index (inner lists) correspond
        to the skymap. Reuse the axes and projections by passing
        ``projections=axs`` to another invocation of ``gridplot``.

    Raises
    ------
    ValueError
        If ``scatters`` is provided and is ill-formatted or not of the same
        length as ``skymaps``; if ``subplot_kwargs`` is included and is not of
        the same length as ``projections``, or if its elements are neither
        ``None`` nor lists of the same length as ``skymaps``; if ``fig`` is
        passed as a ``GridSpec`` which is not compatible with the rest of the
        arguments given; or if one of the
        ``projections`` is specified as a string but cannot be found in this
        module.

    See Also
    --------
    plot
    """

    from math import ceil
    from matplotlib.pyplot import figure
    from matplotlib.figure import Figure
    from matplotlib.gridspec import GridSpec
    from astropy.io.fits import Header
    from astropy.wcs import WCS

    from .partial import PartialUniqSkymap

    gs = fig if isinstance(fig, GridSpec) else None
    fig = fig if gs is None else gs.figure # type: ignore
    projections = list(projections)
    nᵖ = len(projections)  # number of projections per skymap
    nᶜ = len(skymaps)       # number of cells, i.e. skymaps
    nˢ = nᶜ*nᵖ          # number of true subplots in fig
    if gs is None:
        nʳ = nᵖ*ncols       # number of true subplots per row
        nrows = int(ceil(nᶜ/ncols))
    else:
        nʳ = gs.ncols
        if nʳ % nᵖ != 0:
            raise ValueError(f"Cannot evenly fit {nᵖ} projections in "
                             f"{nʳ} columns.")
        ncols = int(nʳ / nᵖ)
        nrows = gs.nrows
        if nrows * nʳ < nˢ:
            raise ValueError(f"Not enough rows {nrows} for {nˢ} subplots.")

    if scatters is None:
        scatters = [
            s.point_sources if isinstance(s, PartialUniqSkymap) else []
            for s in skymaps
        ]
    subplot_kwargs = subplot_kwargs or [None]*nᵖ
    if isinstance(wshrink, Sized):
        if len(wshrink) != nᵖ:
            raise ValueError(f"Must provide one wshrink {wshrink} for "
                             f"each projection {projections}")
    else:
        wshrink = [wshrink]*nᵖ

    if len(scatters) != nᶜ:
        raise ValueError(f"scatters {scatters} must have same len as skymaps "
                         f"{skymaps}")
    if len(subplot_kwargs) != nᵖ:
        raise ValueError(f"subplot_kwargs {subplot_kwargs} must have same "
                         f"len as projections {projections}")

    # initialize kwargs
    for i in range(len(subplot_kwargs)):
        kw = subplot_kwargs[i]
        if kw is None:
            subplot_kwargs[i] = [{}]*nᶜ # type: ignore
        elif len(kw) != nᶜ:
            raise ValueError(f"{i}-th element of subplot_kwargs {kw} must "
                             f"have same len as skymaps {skymaps} or else be "
                             "omitted.")
        del kw
        for j in range(nᶜ):
            subplot_kwargs[i][j].update(kwargs) # type: ignore

    # initialize frame classes and projections
    frames = []
    for iᵖ in range(len(projections)):
        p = projections[iᵖ]
        fa = []
        pa = []
        if isinstance(p, (str, WCS, Header)):
            p = [p]*nᶜ
        for iᶜ, pp in enumerate(p):
            kw = subplot_kwargs[iᵖ][iᶜ] # type: ignore
            if isinstance(pp, (str, WCS, Header)):
                fa.append(get_frame_class(
                    pp, scatter=scatters[iᶜ],
                    **{k: v for k, v in kw.items()
                       if k in GET_FRAME_CLASS_KWARG_KEYS}
                ))
                pa.append(get_projection(
                    pp, scatter=scatters[iᶜ],
                    **{k: v for k, v in kw.items()
                       if k in GET_WCS_KWARG_KEYS}
                ))
            else:
                fa.append(pp.frame_class)
                pa.append(pp.wcs)
        projections[iᵖ] = pa
        frames.append(fa)

    # from https://matplotlib.org/tutorials/intermediate/gridspec.html
    widths = [w * p[0].pixel_shape[0] / p[0].pixel_shape[1]
              for w, p in zip(wshrink, projections)]
    width_ratios = widths*ncols
    nat_width, nat_height = (subplot_height*sum(width_ratios),
                             subplot_height*nrows)
    if not isinstance(fig, Figure):
        fkw = {
            'figsize': (nat_width, nat_height),
            'facecolor': 'w',
            'edgecolor': 'k',
        }
        fkw.update(**(fig or {})) # type: ignore
        fig = figure(**fkw)
    if gs is None:
        gs = fig.add_gridspec(
            nrows=nrows,
            ncols=nᵖ*ncols,
            width_ratios=width_ratios,
            hspace=hspace,
            wspace=wspace,
            left=left,
            right=right,
            bottom=bottom,
            top=top,
        )

    axs = []
    for iᵖ, (p, f, k) in enumerate(zip(projections, frames, subplot_kwargs)):
        axr = []
        for iᶜ, s in enumerate(skymaps):
            row, col = divmod(iᶜ*nᵖ+iᵖ, nʳ)
            ax = plot(s, *scatters[iᶜ], projection=p[iᶜ], fig=fig,
                      subplot=gs[row, col], frame_class=f[iᶜ], **k[iᶜ])
            # hide axis labels
            co_ra, co_dec = ax.coords
            co_ra.set_axislabel_visibility_rule('labels')
            co_dec.set_axislabel_visibility_rule('labels')
            #co_ra.set_ticklabel_visible(row + 1 == nrows)
            #co_dec.set_ticklabel_visible(False)
            axr.append(ax)
        axs.append(axr)

    return gs, axs


def _parametric_ra_exclusion(ra_height, w):
    import numpy as np

    return w.pixel_to_world(
        *np.meshgrid(*map(np.arange, w.pixel_shape), sparse=True)
    ).icrs.ra.value[
        ra_height(w),
        np.concatenate((
            np.arange(int(w.pixel_shape[0])//EDGE_EXCLUSION),
            -np.arange(1, int(w.pixel_shape[0])//EDGE_EXCLUSION),
        ))
    ]


def _dec_exclusion(w):
    import numpy as np

    dec = w.pixel_to_world(
            *np.meshgrid(*map(np.arange, w.pixel_shape), sparse=True)
        ).icrs.dec.value[
        np.concatenate((
            np.arange(int(w.pixel_shape[1])//EDGE_EXCLUSION),
            np.arange(1, int(w.pixel_shape[1])//EDGE_EXCLUSION),
        ))
    ]
    return dec[~np.isnan(dec)]


def _exclusions(registry, ax):
    _init_exclusions()
    w = ax.wcs
    frame_class = ax.coords.frame.__class__
    for c in frame_class.mro():
        if c in registry:
            return registry[c](w)


def register_ra_exclusion(cls, func):
    _init_exclusions()
    _RA_EXCLUSIONS[cls] = func


def register_dec_exclusion(cls, func):
    _init_exclusions()
    _DEC_EXCLUSIONS[cls] = func


# to allow lazy-loading
def _init_exclusions():
    from astropy.visualization.wcsaxes.frame import EllipticalFrame

    if not EXCLUSIONS_INITIALIZED[0]:
        _RA_EXCLUSIONS[EllipticalFrame] = \
            partial(_parametric_ra_exclusion,
                    lambda w: int(w.pixel_shape[1])//2)
        EXCLUSIONS_INITIALIZED[0] = True


_RA_EXCLUSIONS = {
    object: partial(_parametric_ra_exclusion, lambda _: -1,),
}
_DEC_EXCLUSIONS = {
    object: _dec_exclusion,
}
ra_exclusions = partial(_exclusions, _RA_EXCLUSIONS)
ra_exclusions.__doc__ = "Get RA values to exclude from ticks."
dec_exclusions = partial(_exclusions, _DEC_EXCLUSIONS)
dec_exclusions.__doc__ = "Get declination values to exclude from ticks."


def get_ticks(include, exclude, ticks, delta, transform):
    import numpy as np

    bins = np.repeat(ticks, 2).reshape((-1, 2))
    bins += np.array([-1, 1]) * delta
    bins = bins.ravel()
    ihist, _ = np.histogram(transform(include, delta), bins=bins)
    ehist, _ = np.histogram(transform(exclude, delta), bins=bins)
    matches = (ihist != 0) & (ehist == 0)
    return ticks[matches[::2]]
